from typing import List, Optional, Union, Literal
from pydantic import Field
from visivo.models.base.named_model import NamedModel


class BaseDashboard(NamedModel):
    """Base class for all dashboard types (internal and external)"""

    level: Optional[Union[int, str]] = Field(
        None,
        description="The importance level of the dashboard (either an index number or level title)",
    )
    tags: List[str] = Field(
        default_factory=list, description="A list of tags associated with the dashboard"
    )
    description: Optional[str] = Field(
        None, description="A description of the dashboard's purpose and contents"
    )
    type: Literal["internal", "external"] = Field(
        ..., description="The type of dashboard (internal or external)"
    )
    name: str = Field(..., description="The unique name of the dashboard across the entire project")

    def for_each_item(self, function):
        pass



from pydantic import BaseModel
from visivo.models.sources.source import Source
from visivo.models.models.model import Model
from visivo.models.trace import Trace


def list_phase(project, object_type):
    objects = []

    def collect_objects(item, target_type):
        collected = []
        if isinstance(item, list):
            for sub_item in item:
                collected.extend(collect_objects(sub_item, target_type))
        elif isinstance(item, BaseModel):
            for f in item.model_fields_set:
                collected.extend(collect_objects(getattr(item, f), target_type))
            if "name" in item.model_fields_set:
                name = getattr(item, "name")
                if isinstance(item, Source) and target_type == "sources":
                    collected.append(name)
                elif isinstance(item, Model) and target_type == "models":
                    collected.append(name)
                elif isinstance(item, Trace) and target_type == "traces":
                    collected.append(name)

        return collected

    objects = collect_objects(project, object_type)
    objects = list(map(lambda x: f" - {x}", objects))

    # Print each object's name on a new line
    print(f"{object_type}:\n" + "\n".join(objects))



from visivo.models.dag import all_descendants, all_descendants_of_type
from visivo.models.selector import Selector
from visivo.models.sources.source import Source
from visivo.models.project import Project
from visivo.models.chart import Chart
from visivo.models.table import Table
from visivo.models.trace import Trace
from visivo.models.models.model import Model
from visivo.version import VISIVO_VERSION


class Serializer:
    def __init__(self, project: Project):
        self.project = project

    def create_flattened_project(self) -> dict:
        """
        Creates a flattened version of the project where all objects are at the top level
        and nested objects are replaced with references.
        """
        project = self.project
        dag = project.dag()

        # Collect all objects
        all_sources = []
        all_models = []
        all_traces = []
        all_charts = []
        all_tables = []
        all_selectors = []

        # Process all objects in the DAG
        for node in dag.nodes():
            if isinstance(node, Source):
                all_sources.append(node.model_dump(exclude_none=True, mode="json"))
            elif isinstance(node, Model):
                all_models.append(node.model_dump(exclude_none=True, mode="json"))
            elif isinstance(node, Trace):
                all_traces.append(node.model_dump(exclude_none=True, mode="json"))
            elif isinstance(node, Chart):
                all_charts.append(node.model_dump(exclude_none=True, mode="json"))
            elif isinstance(node, Table):
                all_tables.append(node.model_dump(exclude_none=True, mode="json"))
            elif isinstance(node, Selector):
                all_selectors.append(node.model_dump(exclude_none=True, mode="json"))

        # Create the flattened structure
        flattened = {
            "name": project.name,
            "cli_version": project.cli_version,
            "sources": all_sources,
            "models": all_models,
            "traces": all_traces,
            "charts": all_charts,
            "tables": all_tables,
            "selectors": all_selectors,
        }

        # Add default source name if it exists
        if project.defaults and project.defaults.source_name:
            flattened["default_source"] = project.defaults.source_name

        return flattened

    def dereference(self) -> Project:
        project = self.project.model_copy(deep=True)
        project.cli_version = VISIVO_VERSION
        dag = project.dag()

        for dashboard in project.dashboards:

            def replace_item_ref(item):
                if item.chart or item.table:
                    if item.chart:
                        item.chart = all_descendants_of_type(type=Chart, dag=dag, from_node=item)[0]
                        component = item.chart
                    else:
                        item.table = all_descendants_of_type(type=Table, dag=dag, from_node=item)[0]
                        component = item.table

                    component.traces = all_descendants_of_type(
                        type=Trace, dag=dag, from_node=component, depth=1
                    )
                    if component.selector:
                        component.selector = all_descendants_of_type(
                            type=Selector, dag=dag, from_node=component, depth=1
                        )[0]
                        component.selector.options = all_descendants_of_type(
                            type=Trace,
                            dag=dag,
                            from_node=component.selector,
                            depth=1,
                        )
                    for trace in component.traces:
                        trace.model = all_descendants_of_type(type=Model, dag=dag, from_node=trace)[
                            0
                        ]
                        if hasattr(trace.model, "source"):
                            trace.model.source = all_descendants_of_type(
                                type=Source, dag=dag, from_node=trace.model
                            )[0]
                        if hasattr(trace.model, "models"):
                            trace.model.models = all_descendants_of_type(
                                type=Model, dag=dag, from_node=trace.model
                            )
                if item.selector:
                    item.selector = all_descendants_of_type(
                        type=Selector, dag=dag, from_node=item, depth=1
                    )[0]
                    options = [
                        option
                        for option in all_descendants(dag=dag, from_node=item.selector, depth=1)
                        if not isinstance(option, Selector)
                    ]
                    item.selector.options = options

            dashboard.for_each_item(replace_item_ref)

        project.charts = []
        project.traces = []
        project.tables = []
        project.models = []
        project.sources = []
        project.selectors = []
        return project



from textwrap import dedent
import yaml
import re

MARKDOWN_SAFE_DESCRIPTION = "Sets the constraint operation. `=` keeps regions equal to `value`. `<` and `<=` keep regions less than `value`. `>` and `>=` keep regions greater than `value`. `[]`, `()`, `[)`, and `(]` keep regions inside `value[0]` and `value[1]`. `][`, `)(`, `](`, and `)[` keep regions outside `value[0]` and `value[1]`. Open vs. closed intervals make no difference to constraint display, but all versions are allowed for consistency with filter transforms."


def find_refs(obj):
    refs = []
    if isinstance(obj, dict):
        for key, value in obj.items():
            if key == "$ref":
                refs.append(value)
            elif isinstance(value, (dict, list)):
                refs.extend(find_refs(value))
    elif isinstance(obj, list):
        for item in obj:
            if isinstance(item, (dict, list)):
                refs.extend(find_refs(item))
    return sorted(refs)


def handle_allOf(attribute_property_object: dict, model_defs: dict):

    default = attribute_property_object.get("default", "None")
    model_key = find_refs(attribute_property_object)[0].split("/")[-1]
    model = model_defs.get(model_key, {})
    if not model:
        raise KeyError(f"Key {model_key} was not found in $defs dictionary.")

    if model.get("enum"):
        if model.get("type") not in ["string", "object"]:
            enum_options = [str(a) for a in model.get("enum")]
        else:
            enum_options = model.get("enum")
        type = "Enumerated - one of: " + ", ".join(enum_options)
    else:
        type = model.get("type", "None")

    return type, default


def handle_anyOf(attribute_property_object: dict):
    default = attribute_property_object.get("default", "None")
    anyOf = find_refs(attribute_property_object)
    for option in attribute_property_object["anyOf"]:
        if "pattern" in option and find_refs(attribute_property_object) == []:
            anyOf.append(f"Regex({option.get('pattern')})")
        elif "oneOf" in option and len(option.get("oneOf")) == 1:
            anyOf.append(option.get("oneOf")[0].get("type"))
        elif "type" in option and "pattern" not in option:
            if option.get("type") not in ["null", "None"]:
                anyOf.append(option.get("type"))
    if len(anyOf) > 1:
        return "Any of: " + ", ".join(anyOf), default
    elif len(anyOf) == 1:
        return anyOf[0], default


def handle_const(attribute_property_object: dict):
    default = attribute_property_object.get("default", "None")
    return "string", default


def handle_items(attribute_property_object: dict):
    default = attribute_property_object.get("default", "None")
    type = attribute_property_object.get("type")
    refs = find_refs(attribute_property_object.get("items", {}))
    if type == "array" and refs:
        return "Array of " + " or ".join(refs), default
    return type, default


def handle_discriminator(attribute_property_object: dict):
    default_key = attribute_property_object.get("discriminator", {}).get("propertyName", "")
    default_value = attribute_property_object.get("default", {}).get(default_key)
    default = (
        attribute_property_object.get("discriminator", {}).get("mapping", {}).get(default_value)
    )
    refs = find_refs(attribute_property_object.get("oneOf", []))
    if refs:
        return "One of:<br>  •" + "<br>  •".join(refs), default
    return "", default


def handle_default(attribute_property_object: dict):
    default = attribute_property_object.get("default", "None")
    type = attribute_property_object.get("type", "")
    return type, default


def handle_attribute_properties(model_defs: dict, attribute_property_object: dict):
    """Handle extracting type, description, default from different types of property objects that pydantic produces"""
    description = attribute_property_object.get("description", "")

    attribute_key_type = ".".join(list(attribute_property_object.keys()))
    if "allOf" in attribute_key_type:
        try:
            type, default = handle_allOf(attribute_property_object, model_defs)
        except Exception as e:
            raise e
    elif "anyOf" in attribute_key_type:
        type, default = handle_anyOf(attribute_property_object)
    elif "const" in attribute_key_type:
        type, default = handle_const(attribute_property_object)
    elif "items" in attribute_key_type:
        type, default = handle_items(attribute_property_object)
    elif "discriminator" in attribute_key_type:
        type, default = handle_discriminator(attribute_property_object)
    else:
        type, default = handle_default(attribute_property_object)

    default = default if default is not None else "None"

    return type, description, default


def from_pydantic_model(model_defs: dict, model_name: str) -> str:
    """Generates markdown from the model description and creates table of attributes from $defs in pydantic schema"""
    model_def = model_defs.get(model_name, {})
    if not model_def:
        raise KeyError(f"Schema missing model: {model_name}")
    model_properties = model_def.get("properties", {})
    model_md = "" if not model_def.get("description", {}) else dedent(model_def.get("description"))
    md_table = (
        "| Field | Type | Default | Description |\n|-------|------|---------|-------------|\n"
    )

    for property_name, property_object in model_properties.items():
        field_type, field_description, field_default = handle_attribute_properties(
            model_defs, property_object
        )
        md_table += f"| {property_name} | {field_type} | {field_default} | {field_description} |\n"

    return f"# {model_name}" + "\n" + model_md + "\n## Attributes\n" + md_table


def _get_traceprop_nested_structure(model: dict, details: list = []) -> str:
    """Generates Trace Props reference dictionary that will later be converted into yaml for the md file"""
    model_properties = model.get("properties", {})
    if not model_properties:
        raise KeyError(f"Model not found in model_defs dictionary passed into the function.")
    output = {}

    for field_name, field_info in model_properties.items():
        field_info_keys = ".".join(list(field_info.keys()))

        if field_info == {}:
            output[field_name] = "any"
        elif "oneOf" in field_info_keys:
            refs = find_refs(field_info.get("oneOf", {}))

            # Remove query-string and color refs
            refs = [
                ref
                for ref in refs
                if ref not in ["#/$defs/query-string", "#/$defs/color", "#/$defs/colorscale"]
            ]
            if refs and len(refs) == 1:
                nested_model_name = refs[0].split("/")[-1]
                print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
                print(nested_model_name)
                print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")

                # output[field_name], details = _get_traceprop_nested_structure(
                #     model_defs, nested_model_name, details
                # )
            elif refs and len(refs) > 1:
                raise NotImplementedError(
                    "Have not handled Traceprop attributes with multiple models referenced."
                )
            else:
                field_description = field_info.get("description", "")
                if field_description.startswith("Sets the constraint operation."):
                    field_description = MARKDOWN_SAFE_DESCRIPTION
                position = len(details) + 1

                def get_type_from_one_of(one_of):
                    if one_of.get("type", None):
                        return one_of.get("type")
                    if one_of.get("oneOf", None):
                        for sub_one_of in one_of.get("oneOf"):
                            if sub_one_of.get("type", None):
                                return sub_one_of.get("type")
                            if sub_one_of.get("$ref", None):
                                if sub_one_of.get("$ref") == "#/$defs/color":
                                    return "color"
                                if sub_one_of.get("$ref") == "#/$defs/colorscale":
                                    return "colorscale"

                types = list(map(get_type_from_one_of, field_info.get("oneOf", [])))
                # Filter out None types
                types = [t for t in types if t is not None]

                type = " | ".join(types)
                if len(type) == 0:
                    type = "any"
                if len(field_description.strip()) > 0:
                    type = type + f" #({position})!"
                    detail_line = f"{position}. " + field_description
                    details.append(detail_line)
                output[field_name] = type
        elif "const" in field_info_keys:
            output[field_name] = field_info.get("const")
        elif "type" in field_info_keys:
            type = field_info.get("type")
            if type == "object":
                output[field_name], details = _get_traceprop_nested_structure(field_info, details)
            else:
                output[field_name] = field_info.get("type")
                if field_info.get("description"):
                    output[field_name] = (
                        output[field_name] + " #(" + field_info.get("description") + ")"
                    )
                type = field_info.get("type")
                output[field_name] = type
                position = len(details) + 1
                if (
                    field_info.get("description")
                    and len(field_info.get("description", "").strip()) > 0
                ):
                    type = type + f" #({position})!"
                    detail_line = f"{position}. " + field_info.get("description", "")
                    details.append(detail_line)
                output[field_name] = type
        elif "properties" in field_info_keys:
            output[field_name], details = _get_traceprop_nested_structure(field_info, details)
        elif "description" in field_info_keys or "$ref" in field_info_keys:
            pass
        else:
            raise NotImplementedError(
                f"Have not yet handled properties with attributes {field_info_keys} "
            )

    return output, details


def from_traceprop_model(model_defs: dict, model_name: str) -> str:
    model_def = model_defs.get(model_name, {})
    if not model_def:
        raise KeyError(f"Schema missing model: {model_name}")
    model_md = (
        "" if not model_def.get("description", {}) else dedent(model_def.get("description")) + "\n"
    )
    if model_name.lower() == "layout":
        description = "These attributes apply to the `chart.layout` object.\n"
    else:
        description = f"These attributes apply to traces where `trace.props.type` is set to `{model_name.lower()}`. You would configure these attributes on the trace with the `trace.props` object.\n"

    model = model_defs.get(model_name)
    nested_structure, details = _get_traceprop_nested_structure(model, details=[])
    if model_name.lower() == "scatter":
        title = "Scatter (line, area & scatter)"
    else:
        title = model_name
    yaml_doc = yaml.dump(nested_structure, default_flow_style=False)
    pattern = r"'([^'#]+) (\#\(.*?\)!)'"
    processed_yaml_doc = re.sub(pattern, r"'\1' \2", yaml_doc)
    if model_name.lower() != "layout":
        includes = (
            "{!"
            + f" include-markdown '"
            + f"reference/props-docs/{model_name.lower()}.md' "
            + f"\nstart='<!--start-->'"
            + f"\nend='<!--end-->'"
            + "!}"
        )
    else:
        includes = ""
    full_doc = (
        f"# {title} "
        + "\n"
        + includes
        + model_md
        + "{% raw %}\n"
        + "## Attributes\n"
        + description
        + "``` yaml\n"
        + processed_yaml_doc
        + "\n```\n\n"
        + "\n".join(details)
        + "\n{% endraw %}\n"
    )
    return full_doc



import shutil
from pathlib import Path


def copy_script():
    project_root = Path(__file__).parent.parent
    install_script_src = project_root / "install.sh"
    install_script_dest = project_root / "mkdocs" / "assets" / "install.sh"

    print(f"Copying {install_script_src} to {install_script_dest}")
    shutil.copy(install_script_src, install_script_dest)


if __name__ == "__main__":
    copy_script()



from visivo.server.models.base import Base, Column, String, DateTime, relationship
from datetime import datetime


class WorksheetModel(Base):
    __tablename__ = "worksheets"

    id = Column(String, primary_key=True)
    name = Column(String, nullable=False)
    query = Column(String)
    selected_source = Column(String)
    created_at = Column(DateTime, default=datetime.utcnow)
    updated_at = Column(DateTime, default=datetime.utcnow, onupdate=datetime.utcnow)
    last_run_at = Column(DateTime)

    session_state = relationship(
        "SessionStateModel", back_populates="worksheet", uselist=False, cascade="all, delete-orphan"
    )
    results = relationship("ResultModel", back_populates="worksheet", cascade="all, delete-orphan")

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "query": self.query,
            "selected_source": self.selected_source,
            "created_at": self.created_at.isoformat() if self.created_at else None,
            "updated_at": self.updated_at.isoformat() if self.updated_at else None,
            "last_run_at": self.last_run_at.isoformat() if self.last_run_at else None,
        }



import os
import yaml
import click
from visivo.logger.logger import Logger

PROFILE_PATH = os.path.expanduser("~/.visivo/profile.yml")


def get_existing_token():
    """Check if a token exists in the profile file."""
    if not os.path.exists(PROFILE_PATH):
        return None
    with open(PROFILE_PATH, "r") as profile_file:
        profile_yaml = yaml.safe_load(profile_file) or {}
        return profile_yaml.get("token")


def write_token(token):
    """Write the token to the profile file."""
    profile_dir = os.path.dirname(PROFILE_PATH)
    os.makedirs(profile_dir, exist_ok=True)
    try:
        with open(PROFILE_PATH, "w") as f:
            yaml.dump({"token": token}, f)
        Logger.instance().success("Token written successfully")
    except Exception as e:
        raise click.ClickException(f"Error writing token to {PROFILE_PATH}: {str(e)}")


def validate_and_store_token(token):
    """Prompt for a token if none provided, validate length, then write it."""
    if token is None:
        token = click.prompt("Please enter your token")

    if len(token) < 10:
        raise click.ClickException(
            "Token is too short. Please ensure you have the correct token "
            "from the Visio Settings Profile Page"
        )

    Logger.instance().debug("Token received")
    Logger.instance().debug(f"Writing token to {PROFILE_PATH}")
    write_token(token)
    Logger.instance().debug("Token updated successfully")



from visivo.discovery.discover import Discover
from visivo.parsers.parser_factory import ParserFactory
from visivo.models.project import Defaults
from visivo.logger.logger import Logger
from visivo.commands.dbt_phase import dbt_phase
from visivo.utils import get_dashboards_dir
import yaml
import click
import os
from time import time


def parse_project_phase(working_dir, output_dir, default_source, dbt_profile, dbt_target):

    # Run and Track dbt phase
    dbt_start = time()
    Logger.instance().debug("    Running dbt phase...")
    dbt_phase(working_dir, output_dir, dbt_profile, dbt_target)
    dbt_duration = round(time() - dbt_start, 2)
    if os.environ.get("STACKTRACE"):
        Logger.instance().info(f"dbt phase completed in {dbt_duration}s")

    discover = Discover(working_dir=working_dir, output_dir=output_dir)
    parser = ParserFactory().build(project_file=discover.project_file, files=discover.files)

    project = None
    try:
        project = parser.parse()
        if not project.defaults:
            project.defaults = Defaults()
        if default_source:
            project.defaults.source_name = default_source
        # Ensure output directory exists
        os.makedirs(output_dir, exist_ok=True)
        # Ensure dashboard directory exists
        dashboard_dir = get_dashboards_dir(output_dir)
        os.makedirs(dashboard_dir, exist_ok=True)

    except yaml.YAMLError as e:
        message = "\n"
        if hasattr(e, "problem_mark"):
            mark = e.problem_mark
            message = f"\n Error position: line:{mark.line+1} column:{mark.column+1}\n"
        raise click.ClickException(f"There was an error parsing the yml file(s):{message} {e}")

    return project



from pydantic import Field, HttpUrl
from visivo.models.dashboards.base_dashboard import BaseDashboard
from typing import Literal


class ExternalDashboard(BaseDashboard):
    """
    Represents an external dashboard that links to another system or URL
    """

    type: Literal["external"] = Field(
        "external", description="The type of dashboard (always 'external')"
    )
    href: HttpUrl = Field(..., description="URL to the external dashboard")



from visivo.models.models.local_merge_model import LocalMergeModel
from visivo.jobs.job import (
    Job,
    JobResult,
    format_message_failure,
    format_message_success,
    start_message,
)
from time import time
from visivo.logger.logger import Logger


def action(local_merge_model: LocalMergeModel, output_dir, dag):
    Logger.instance().info(start_message("LocalMergeModel", local_merge_model))
    try:
        start_time = time()
        local_merge_model.insert_duckdb_data(output_dir=output_dir, dag=dag)
        success_message = format_message_success(
            details=f"Updated data for model \033[4m{local_merge_model.name}\033[0m",
            start_time=start_time,
            full_path=local_merge_model.get_duckdb_source(output_dir=output_dir, dag=dag).database,
        )
        return JobResult(item=local_merge_model, success=True, message=success_message)
    except Exception as e:
        failure_message = format_message_failure(
            details=f"Failed query for model \033[4m{local_merge_model.name}\033[0m",
            start_time=start_time,
            full_path=local_merge_model.get_duckdb_source(output_dir=output_dir, dag=dag).database,
            error_msg=str(repr(e)),
        )
        return JobResult(item=local_merge_model, success=False, message=failure_message)


def job(dag, output_dir: str, local_merge_model: LocalMergeModel):
    return Job(
        item=local_merge_model,
        source=local_merge_model.get_duckdb_source(output_dir=output_dir, dag=dag),
        action=action,
        local_merge_model=local_merge_model,
        output_dir=output_dir,
        dag=dag,  # PR question: Why are we passing the dag here to local merge, trace models but not csv script models? It seems like it gets what it needs from the source pass no?
    )



import urllib.parse
import webbrowser


def open_url(url: str) -> bool:
    """
    Tries to open the given URL in the default web browser.
    Returns True if webbrowser.open() claims success.
    """
    parsed = urllib.parse.urlparse(url)
    query_dict = dict(urllib.parse.parse_qsl(parsed.query))
    encoded_query = urllib.parse.urlencode(query_dict)
    safe_url = urllib.parse.urlunparse(
        (
            parsed.scheme,
            parsed.netloc,
            parsed.path,
            parsed.params,
            encoded_query,
            parsed.fragment,
        )
    )
    return webbrowser.open(safe_url)


def generate_success_html_response(base_url: str, timeout: int = 5) -> str:
    """
    Returns an HTML page that will redirect to base_url + '/profile' after `timeout` seconds.
    """
    redirect_url = f"{base_url}/profile"
    return f"""
    <html>
      <head>
        <meta http-equiv="refresh" content="{timeout};url={redirect_url}" />
        <title>Authorization Successful</title>
        <style>
          body {{
            display: flex;
            align-items: center;
            justify-content: center;
            height: 100vh;
            margin: 0;
            font-family: Arial, sans-serif;
            background-color: #f0f0f0;
          }}
          h1 {{
            font-size: 3em;
            font-weight: bold;
            text-align: center;
            color: #333;
          }}
          p {{
            text-align: center;
            font-size: 1.2em;
            color: #666;
          }}
        </style>
      </head>
      <body>
        <div>
          <h1>A token has been written to your development machine to this file ~/.visivo/profile.yml</h1>
          <p>You'll now be redirected to your profile page with more instructions on
          how to create a new stage and project should this be necessary.</p>
        </div>
      </body>
    </html>
    """



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session
from visivo.server.models.worksheet import Base, WorksheetModel
from visivo.server.models.session_state import SessionStateModel
from visivo.server.models.result import ResultModel
from datetime import datetime
import uuid


class WorksheetRepository:
    def __init__(self, db_path: str):
        """Initialize the repository with a database path."""
        self.engine = create_engine(f"sqlite:///{db_path}")
        Base.metadata.create_all(self.engine)
        session_factory = sessionmaker(bind=self.engine)
        self.Session = scoped_session(session_factory)

    def create_worksheet(self, name: str, query: str = "", selected_source: str = None):
        """Create a new worksheet with session state."""
        session = self.Session()
        try:
            # Get max tab order
            max_order = (
                session.query(SessionStateModel.tab_order)
                .order_by(SessionStateModel.tab_order.desc())
                .first()
            )
            next_order = (max_order[0] if max_order else 0) + 1

            # Create worksheet
            worksheet = WorksheetModel(
                id=str(uuid.uuid4()), name=name, query=query, selected_source=selected_source
            )

            # Create session state
            worksheet.session_state = SessionStateModel(tab_order=next_order, is_visible=True)

            session.add(worksheet)
            session.commit()
            return {
                "worksheet": worksheet.to_dict(),
                "session_state": worksheet.session_state.to_dict(),
            }
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def get_worksheet(self, worksheet_id: str):
        """Get a worksheet by ID with its latest result."""
        session = self.Session()
        try:
            worksheet = session.query(WorksheetModel).filter_by(id=worksheet_id).first()
            if not worksheet:
                return None

            latest_result = (
                session.query(ResultModel)
                .filter_by(worksheet_id=worksheet_id)
                .order_by(ResultModel.created_at.desc())
                .first()
            )

            return {
                "worksheet": worksheet.to_dict(),
                "session_state": worksheet.session_state.to_dict(),
                "results": latest_result.to_dict() if latest_result else None,
            }
        finally:
            session.close()

    def list_worksheets(self):
        """List all worksheets ordered by tab order."""
        session = self.Session()
        try:
            worksheets = (
                session.query(WorksheetModel)
                .join(SessionStateModel)
                .order_by(SessionStateModel.tab_order)
                .all()
            )
            return [
                {"worksheet": w.to_dict(), "session_state": w.session_state.to_dict()}
                for w in worksheets
            ]
        finally:
            session.close()

    def update_worksheet(self, worksheet_id: str, updates: dict):
        """Update a worksheet's attributes."""
        session = self.Session()
        try:
            worksheet = session.query(WorksheetModel).filter_by(id=worksheet_id).first()
            if not worksheet:
                return False

            valid_fields = {"name", "query", "selected_source"}
            for key, value in updates.items():
                if key in valid_fields:
                    setattr(worksheet, key, value)

            session.commit()
            return True
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def update_session_states(self, states: list):
        """Update multiple session states in a transaction."""
        session = self.Session()
        try:
            for state in states:
                session_state = (
                    session.query(SessionStateModel)
                    .filter_by(worksheet_id=state["worksheet_id"])
                    .first()
                )
                if session_state:
                    session_state.tab_order = state["tab_order"]
                    session_state.is_visible = state["is_visible"]
            session.commit()
            return True
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def save_results(self, worksheet_id: str, results_json: str, query_stats_json: str):
        """Save query results for a worksheet."""
        session = self.Session()
        try:
            worksheet = session.query(WorksheetModel).filter_by(id=worksheet_id).first()
            if not worksheet:
                return False

            worksheet.last_run_at = datetime.utcnow()
            result = ResultModel(
                worksheet=worksheet, results_json=results_json, query_stats_json=query_stats_json
            )
            session.add(result)
            session.commit()
            return True
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def delete_worksheet(self, worksheet_id: str):
        """Delete a worksheet and its associated data."""
        session = self.Session()
        try:
            worksheet = session.query(WorksheetModel).filter_by(id=worksheet_id).first()
            if not worksheet:
                return False
            session.delete(worksheet)
            session.commit()
            return True
        except:
            session.rollback()
            raise
        finally:
            session.close()



from visivo.models.dashboard import Dashboard
from visivo.models.project import Project
import os


def run_phase(
    default_source: str,
    output_dir: str,
    working_dir: str,
    dag_filter: str = None,
    threads: int = None,
    soft_failure: bool = False,
    dbt_profile: str = None,
    dbt_target: str = None,
    skip_compile: bool = False,
    project: Project = None,
    server_url: str = None,
):
    from visivo.logger.logger import Logger
    from visivo.jobs.filtered_runner import FilteredRunner
    from time import time

    # Replace compile phase with parse project phase if skip_compile is True. Injects the project if it's available.
    if project and skip_compile:
        Logger.instance().debug(
            f"Using provided project {project.name}. skip_compile is {skip_compile}"
        )
    elif not project and skip_compile:
        from visivo.commands.parse_project_phase import parse_project_phase

        Logger.instance().info("Parsing project...")
        start_time = time()
        project = parse_project_phase(
            working_dir=working_dir,
            output_dir=output_dir,
            default_source=default_source,
            dbt_profile=dbt_profile,
            dbt_target=dbt_target,
        )
        Logger.instance().info(f"Parsing project took {round(time() - start_time, 2)}s")
    else:
        from visivo.commands.compile_phase import compile_phase

        project = compile_phase(
            default_source=default_source,
            working_dir=working_dir,
            output_dir=output_dir,
            dbt_profile=dbt_profile,
            dbt_target=dbt_target,
        )

    if not dag_filter:
        dag_filter = ",".join(
            map(
                lambda x: f"+{x.name}+",
                project.dag().get_nodes_by_types([Dashboard], True),
            )
        )

    Logger.instance().debug(f"DAG filter: {dag_filter}")
    # Initialize project defaults if not present
    if threads is None and project.defaults and project.defaults.threads:
        threads = project.defaults.threads
    else:
        threads = int(threads)

    source_details = "\n" if default_source == None else f" and default source {default_source}\n"
    Logger.instance().info(f"Running project across {threads} threads" + source_details)

    runner = FilteredRunner(
        project=project,
        output_dir=output_dir,
        threads=threads,
        soft_failure=soft_failure,
        dag_filter=dag_filter,
        server_url=server_url,
    )
    runner.run()
    return runner



import click
from visivo.commands.options import working_dir, output_dir, source
from visivo.commands.parse_project_phase import parse_project_phase
from visivo.commands.list_phase import list_phase


@click.command()
@click.argument("object_type", type=click.Choice(["sources", "models", "traces"]))
@working_dir
@output_dir
@source
def list(object_type, working_dir, output_dir, source):
    """
    Lists all objects of a given type in the project.
    """
    from visivo.logger.logger import Logger

    Logger.instance().debug(f"Listing {object_type}")

    # Parse the project to get access to all objects
    project = parse_project_phase(
        working_dir=working_dir,
        output_dir=output_dir,
        default_source=source,
        dbt_profile=None,
        dbt_target=None,
    )

    list_phase(project, object_type)

    Logger.instance().success("Done")



from typing import Any, List
import re
from visivo.models.base.context_string import ContextString
from visivo.models.test_run import TestRun
from visivo.parsers.evaluator import evaluate_expression

INLINE_CONTEXT_STRING_REGEX = r"\${\s*[\(a-zA-Z0-9\s'\"\-_\\.\]\[)]+?\s*}"
EVAL_STRING_REGEX = r"^>{(.*)}$"

"""
EvalString represents a string that contains expressions to be evaluated.

This class is used for handling strings that may contain references or expressions
enclosed in '>{' and '}' delimiters. It provides functionality to validate the
string format and extract embedded ContextString instances.

Available injected methods:
    - any_test_failed() -> bool

Available injected attributes:
    - env -> Dict[str, Any]
"""


class EvalString:
    def __init__(self, value: str):
        self.value = value

    def __str__(self):
        return self.value

    def get_context_strings(self) -> List[ContextString]:
        return list(
            map(
                lambda m: ContextString(m),
                re.findall(INLINE_CONTEXT_STRING_REGEX, self.value),
            )
        )

    def get_references(self) -> List[str]:
        return list(filter(None, map(lambda c: c.get_reference(), self.get_context_strings())))

    def get_paths(self) -> List[str]:
        return list(filter(None, map(lambda c: c.get_path(), self.get_context_strings())))

    def evaluate(self, dag: Any, project: Any, output_dir: str, test_run: TestRun = None) -> Any:
        expression = re.match(EVAL_STRING_REGEX, self.value.strip()).group(1).strip()
        expression = self.__replace_context_strings(expression=expression, dag=dag)
        return evaluate_expression(
            expression=expression,
            project=project,
            output_dir=output_dir,
            test_run=test_run,
        )

    def __replace_context_strings(self, expression: str, dag: Any) -> str:
        context_strings = self.get_context_strings()
        for context_string in context_strings:
            if context_string.get_reference() is not None:
                item = context_string.get_item(dag=dag)

                path = item.path
                props_path = context_string.get_ref_props_path()
                if props_path is not None:
                    path = path + props_path
                expression = expression.replace(context_string.value, path)
            else:
                expression = expression.replace(context_string.value, context_string.get_path())
        return expression

    @classmethod
    def __get_pydantic_core_schema__(cls, _source_type: Any, handler: Any):
        from pydantic_core import core_schema

        def validate_and_create(value: Any) -> "EvalString":
            if isinstance(value, cls):
                return value
            str_value = str(value)
            if not (str_value.startswith(">{") and str_value.endswith("}")):
                raise ValueError("EvalString must start with '>{' and end with '}'")
            return cls(str_value)

        return core_schema.no_info_after_validator_function(
            validate_and_create,
            core_schema.str_schema(),
            serialization=core_schema.plain_serializer_function_ser_schema(str),
        )



from typing import Any

import re


QUERY_STRING_VALUE_REGEX = r"^\?\{\s*(?P<query_string>.+)\s*\}\s*$"


class QueryString:
    """
    Adds the value of the query string to the query.
    This allows you to reference the output of the query in your chart

    ?{ x }
    """

    def __init__(self, value: str):
        self.value = value

    def __str__(self):
        return self.value

    def get_value(self) -> str:
        matches = re.findall(QUERY_STRING_VALUE_REGEX, self.value)
        if len(matches) == 0:
            return None
        else:
            return matches[0].strip()

    @classmethod
    def __get_pydantic_core_schema__(cls, _source_type: Any, handler: Any):
        from pydantic_core import core_schema

        def validate_and_create(value: Any) -> "QueryString":
            if isinstance(value, cls):
                return value
            str_value = str(value)
            if not (str_value.startswith("?{") and str_value.endswith("}")):
                raise ValueError("QueryString must start with '?{' and end with '}'")
            return cls(str_value)

        return core_schema.no_info_after_validator_function(
            validate_and_create,
            core_schema.str_schema(),
            serialization=core_schema.plain_serializer_function_ser_schema(str),
        )



from visivo.commands.compile_phase import compile_phase
from visivo.logger.logger import Logger

from visivo.server.hot_reload_server import HotReloadServer
from visivo.server.flask_app import FlaskApp
from visivo.commands.run_phase import run_phase
import json
import os
import traceback


def serve_phase(
    output_dir,
    working_dir,
    default_source,
    dag_filter,
    threads,
    skip_compile,
    project,
    server_url,
):

    app = FlaskApp(
        output_dir=output_dir,
        project=project,
    )

    def on_project_change(one_shot=False):
        try:
            Logger.instance().info(
                "Server has detected changes to the project. Re-running project..."
            )
            project = compile_phase(
                default_source=default_source,
                working_dir=working_dir,
                output_dir=output_dir,
            )

            changed_dag_filter = project.dag().get_diff_dag_filter(
                existing_project=app.project, existing_dag_filter=dag_filter
            )
            if not changed_dag_filter:
                Logger.instance().info("No changes to the project.")
                return

            runner = run_phase(
                output_dir=output_dir,
                working_dir=working_dir,
                default_source=default_source,
                dag_filter=changed_dag_filter,
                threads=threads,
                soft_failure=True,
                skip_compile=True,  # Don't recompile on changes, passing in just compiled one.
                project=project,
                server_url=server_url,
            )
            app.project = runner.project
            if one_shot:
                Logger.instance().success("Closing server...")
            else:
                Logger.instance().success("File Change Data Refresh Complete.")
                Logger.instance().info("View your project at: " + server_url)
            with open(f"{output_dir}/error.json", "w") as error_file:
                error_file.write(json.dumps({}))
        except Exception as e:
            error_message = str(e)
            if os.environ.get("STACKTRACE"):
                error_message = f"{str(e)}\n{traceback.format_exc()}"
            Logger.instance().error(error_message)
            with open(f"{output_dir}/error.json", "w") as error_file:
                error_file.write(json.dumps({"message": error_message}))

    def on_server_ready(one_shot=False):
        """Run initialization jobs after server starts"""
        try:
            if one_shot:
                Logger.instance().info("Running project build...")
            else:
                Logger.instance().info("Running initial project build...")
            run_phase(
                output_dir=output_dir,
                working_dir=working_dir,
                default_source=default_source,
                dag_filter=dag_filter,
                threads=threads,
                skip_compile=skip_compile,
                project=project,
                server_url=server_url,
            )
            if one_shot:
                Logger.instance().info("Closing server...")
            else:
                Logger.instance().success("Initial Data Refresh Complete.")
                Logger.instance().info("View your project at: " + server_url)
        except Exception as e:
            error_message = str(e)
            if os.environ.get("STACKTRACE"):
                error_message = f"{str(e)}\n{traceback.format_exc()}"
            Logger.instance().error(error_message)
            raise

    # Create ignore patterns for dbt files if needed
    ignore_patterns = []
    if project.dbt and project.dbt.enabled:
        dbt_file = project.dbt.get_output_file(output_dir, working_dir)
        if dbt_file:
            ignore_patterns.append(dbt_file)

    # Create and return the hot reload server
    server = HotReloadServer(app=app.app, watch_path=working_dir, ignore_patterns=ignore_patterns)

    return server, on_project_change, on_server_ready



from typing import Any
import warnings

from visivo.models.base.parent_model import ParentModel
from visivo.models.models.csv_script_model import CsvScriptModel
from visivo.models.models.local_merge_model import LocalMergeModel
from visivo.models.dashboard import Dashboard
from visivo.models.project import Project
from visivo.logger.logger import Logger
from time import time
from concurrent.futures import Future, ThreadPoolExecutor
import queue
import sys
from visivo.models.sources.source import Source
from visivo.models.trace import Trace
from visivo.jobs.job import JobResult

from visivo.jobs.run_csv_script_job import job as csv_script_job
from visivo.jobs.run_trace_job import job as trace_job
from visivo.jobs.run_local_merge_job import job as local_merge_job
from visivo.jobs.run_source_connection_job import job as source_connection_job
from visivo.jobs.job_tracker import JobTracker
from threading import Lock

warnings.filterwarnings("ignore")


class DagRunner:
    def __init__(
        self,
        project: Project,
        output_dir: str,
        threads: int,
        soft_failure: bool,
        server_url: str,
        job_dag: Any,
    ):
        self.project = project
        self.output_dir = output_dir
        self.threads = threads
        self.soft_failure = soft_failure
        self.server_url = server_url
        self.job_dag = job_dag
        self.job_tracking_dag = job_dag.copy()
        self.project_dag = project.dag()
        self.failed_job_results = []
        self.successful_job_results = []
        self.lock = Lock()

    def run(self):
        complete = False
        job_tracker = JobTracker()
        start_time = time()
        with ThreadPoolExecutor(max_workers=self.threads) as executor:
            while True:
                complete = self.update_job_queue(job_tracker)
                if complete:
                    break

                try:
                    job = job_tracker.get_next_job()
                except queue.Empty:
                    continue

                if job.done() or job.running():
                    pass
                else:
                    job.set_future(executor.submit(job.action, **job.kwargs))
                    job.future.add_done_callback(self.job_callback)

        if len(self.failed_job_results) > 0:
            Logger.instance().info("")
            Logger.instance().info("")
            Logger.instance().error(
                f"\n\nRun failed in {round(time()-start_time, 2)}s with {len(self.failed_job_results)} query error(s)."
            )
            for result in self.failed_job_results:
                Logger.instance().error(str(result.message))
            if not self.soft_failure:
                sys.exit(1)
        elif len(self.successful_job_results) == 0 and len(self.failed_job_results) == 0:
            Logger.instance().error(
                f"\nNo jobs run. Ensure your filter contains nodes that are runnable."
            )
        else:
            Logger.instance().info(f"\nRun finished in {round(time()-start_time, 2)}s")

    def job_callback(self, future: Future):
        job_result: JobResult = future.result(timeout=30)
        if job_result.success:
            Logger.instance().success(str(job_result.message))
            self.successful_job_results.append(job_result)
        else:
            Logger.instance().error(str(job_result.message))
            self.failed_job_results.append(job_result)

    def update_job_queue(self, job_tracker: JobTracker):
        from networkx import descendants as node_descendants

        with self.lock:
            terminal_nodes = [
                n
                for n in self.job_tracking_dag.nodes()
                if self.job_tracking_dag.out_degree(n) == 0 and n != self.project
            ]

            for terminal_node in terminal_nodes:
                descendants = node_descendants(self.job_dag, terminal_node)
                if job_tracker.is_job_name_done(terminal_node.name):
                    self.job_tracking_dag.remove_node(terminal_node)
                    continue
                elif any(
                    job_tracker.is_job_name_failed(descendant.name) for descendant in descendants
                ):
                    Logger.instance().info(
                        f"Skipping job for '{terminal_node.name}' because it has a failed dependency"
                    )
                    self.job_tracking_dag.remove_node(terminal_node)
                    continue

                job = self.create_jobs_from_item(terminal_node)
                if not job:
                    self.job_tracking_dag.remove_node(terminal_node)
                    continue
                if not job_tracker.is_job_name_enqueued(job.name):
                    job_tracker.track_job(job)

            return len(self.job_tracking_dag.nodes()) == 1

    def create_jobs_from_item(self, item: ParentModel):
        if isinstance(item, Trace):
            return trace_job(trace=item, output_dir=self.output_dir, dag=self.project_dag)
        elif isinstance(item, CsvScriptModel):
            return csv_script_job(csv_script_model=item, output_dir=self.output_dir)
        elif isinstance(item, LocalMergeModel):
            return local_merge_job(
                local_merge_model=item, output_dir=self.output_dir, dag=self.project_dag
            )
        elif isinstance(item, Source):
            return source_connection_job(source=item)
        return None



from time import time

compile_import_start = time()
from visivo.logger.logger import Logger

Logger.instance().debug("Compiling project...")
import json

from visivo.parsers.serializer import Serializer
from visivo.commands.parse_project_phase import parse_project_phase

import_duration = round(time() - compile_import_start, 2)
Logger.instance().debug(f"Compile Import completed in {import_duration}s")


def compile_phase(
    default_source: str,
    working_dir: str,
    output_dir: str,
    dbt_profile: str = None,
    dbt_target: str = None,
):
    # Track parse project
    parse_start = time()
    Logger.instance().debug("    Running parse project phase...")
    project = parse_project_phase(working_dir, output_dir, default_source, dbt_profile, dbt_target)
    parse_duration = round(time() - parse_start, 2)
    Logger.instance().debug(f"Project parsing completed in {parse_duration}s")

    # Track artifacts writing
    artifacts_start = time()
    Logger.instance().debug("    Writing artifacts...")

    # Write the original project.json
    with open(f"{output_dir}/project.json", "w") as fp:
        serializer = Serializer(project=project)
        fp.write(serializer.dereference().model_dump_json(exclude_none=True))

    # Write the flattened explorer.json for the QueryExplorer
    with open(f"{output_dir}/explorer.json", "w") as fp:
        serializer = Serializer(project=project)
        explorer_data = serializer.create_flattened_project()
        json.dump(explorer_data, fp)
    artifacts_duration = round(time() - artifacts_start, 2)
    Logger.instance().debug(f"Project artifacts written in {artifacts_duration}s")

    total_duration = round(time() - parse_start, 2)

    with open(f"{output_dir}/error.json", "w") as error_file:
        error_file.write(json.dumps({}))

    Logger.instance().success(
        f"Compile completed in {total_duration}s "
        f"imports: {import_duration}s, "
        f"parse: {parse_duration}s, "
        f"artifacts: {artifacts_duration}s, "
    )

    return project



import click
from visivo.commands.options import (
    working_dir,
    output_dir,
    dag_filter,
    source,
    threads,
    dbt_profile,
    dbt_target,
    skip_compile,
    port,
)


@click.command()
@working_dir
@output_dir
@dag_filter
@source
@threads
@dbt_profile
@dbt_target
@skip_compile
@port
def run(
    output_dir,
    working_dir,
    source,
    dag_filter,
    threads,
    dbt_profile,
    dbt_target,
    skip_compile,
    port,
):
    """
    Compiles the project and then runs the trace queries to fetch data to populate in the traces. Writes all data to the source directory. Can skip the compile with the --skip-compile flag.
    """
    from visivo.logger.logger import Logger
    from visivo.commands.parse_project_phase import parse_project_phase
    from visivo.commands.serve_phase import serve_phase

    Logger.instance().debug("Running")

    # Parse project first
    project = parse_project_phase(
        working_dir=working_dir,
        output_dir=output_dir,
        default_source=source,
        dbt_profile=dbt_profile,
        dbt_target=dbt_target,
    )

    from visivo.commands.run_phase import run_phase

    run_phase(
        default_source=source,
        output_dir=output_dir,
        working_dir=working_dir,
        dag_filter=dag_filter,
        threads=threads,
        dbt_profile=dbt_profile,
        dbt_target=dbt_target,
        skip_compile=skip_compile,
        project=project,
    )

    Logger.instance().success("Done")



from typing import Literal, Optional, List
from visivo.models.base.base_model import BaseModel
from visivo.models.sources.source import Source
from pydantic import Field
import click
import duckdb
import polars as pl

DuckdbType = Literal["duckdb"]


class DuckdbAttachment(BaseModel):
    schema_name: str = Field("Name of the schema to attach the source under.")
    source: "DuckdbSource" = Field(
        None,
        description="Local Duckdb database source to attach in the connection that will be available in the base SQL query.",
    )


class DuckdbSource(Source):
    """
    DuckdbSources hold the connection information to DuckDB data sources.

    !!! example

        === "Simple"

            ``` yaml
                sources:
                  - name: duckdb_source
                    type: duckdb
                    database: local/file/database.db
                    connection_pool_size: 2
            ```

    !!! note

        Recommended environment variable use is covered in the [sources overview.](/topics/sources/)
    """

    type: DuckdbType
    connection_pool_size: Optional[int] = Field(
        1, description="The pool size that is used for this connection."
    )
    attach: Optional[List[DuckdbAttachment]] = Field(
        None,
        description="List of other local Duckdb database sources to attach in the connection that will be available in the base SQL query.",
    )

    def get_connection(self, read_only: bool = False):
        try:

            connection = duckdb.connect(self.database, read_only=read_only)

            if self.attach:
                for attachment in self.attach:
                    connection.execute(
                        f"ATTACH DATABASE '{attachment.source.database}' AS {attachment.schema_name}"
                    )

            return connection

        except Exception as err:
            raise click.ClickException(
                f"Error connecting to source '{self.name}'. Ensure the database exists and the connection properties are correct. Full Error: {str(err)}"
            )

    def read_sql(self, query: str):
        try:
            with self.connect(read_only=True) as connection:
                result = connection.execute(query).pl()
                return result
        except Exception as err:
            raise click.ClickException(f"Error executing query on source '{self.name}': {str(err)}")

    def connect(self, read_only: bool = False):
        return DuckDBConnection(source=self, read_only=read_only)

    def get_dialect(self):
        return "duckdb"


class DuckDBConnection:
    def __init__(self, source: DuckdbSource, read_only: bool = False):
        self.source = source
        self.conn = None
        self.read_only = read_only

    def __enter__(self):
        self.conn = self.source.get_connection(read_only=self.read_only)
        return self.conn

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.conn:
            self.conn.close()
            self.conn = None



import click
from time import time
import json

from visivo.commands.options import (
    dag_filter,
    output_dir,
    working_dir,
    source,
    port,
    threads,
    dbt_profile,
    dbt_target,
    skip_compile,
)


@click.command()
@source
@working_dir
@output_dir
@dag_filter
@port
@threads
@skip_compile
@dbt_profile
@dbt_target
def serve(
    output_dir,
    working_dir,
    source,
    port,
    dag_filter,
    threads,
    skip_compile,
    dbt_profile,
    dbt_target,
):
    """
    Enables fast local development by spinning up a localhost server to run and view your project locally. Visivo will automatically refresh your project and re-run traces that have changed when you make updates to project files.
    """
    start_time = time()
    from visivo.commands.serve_phase import serve_phase
    from visivo.commands.parse_project_phase import parse_project_phase
    from visivo.logger.logger import Logger

    server_url = f"http://localhost:{port}"

    # Parse project first
    project = parse_project_phase(
        working_dir=working_dir,
        output_dir=output_dir,
        default_source=source,
        dbt_profile=dbt_profile,
        dbt_target=dbt_target,
    )

    # Create and configure server & callbacks
    server, on_project_change, on_server_ready = serve_phase(
        output_dir=output_dir,
        working_dir=working_dir,
        default_source=source,
        dag_filter=dag_filter,
        threads=threads,
        skip_compile=skip_compile,
        project=project,
        server_url=server_url,
    )

    # Start serving with hot reload
    serve_duration = time() - start_time
    Logger.instance().info(f"Initial build completed in {round(serve_duration, 2)}s")
    Logger.instance().info(f"Server running at {server_url}")

    # Start the server with file watching
    server.serve(
        host="0.0.0.0",
        port=port,
        on_change_callback=on_project_change,
        on_server_ready=on_server_ready,
    )



import json
import os
import ruamel.yaml
from copy import deepcopy

DELETE = object()


class SubDiff(dict):
    pass


def diff(old_dict, new_dict):
    updates = {}
    for key in new_dict:
        if key not in old_dict:
            updates[key] = new_dict[key]
        elif isinstance(new_dict[key], dict) and isinstance(old_dict[key], dict):
            sub_diff = diff(old_dict[key], new_dict[key])
            if sub_diff:
                updates[key] = SubDiff(sub_diff)
        elif isinstance(new_dict[key], list) and isinstance(old_dict[key], list):
            if old_dict[key] != new_dict[key]:
                updates[key] = new_dict[key]
        else:
            if old_dict[key] != new_dict[key]:
                updates[key] = new_dict[key]
    for key in old_dict:
        if key not in new_dict:
            updates[key] = DELETE
    return updates


# Apply diff
def apply_diff(target, diff):
    for key, value in diff.items():
        if value is DELETE:
            if key in target:
                del target[key]
        elif isinstance(value, SubDiff):
            if key in target and isinstance(target[key], dict):
                apply_diff(target[key], value)
            else:
                target[key] = value
        else:
            target[key] = value


class ProjectWriter:
    """
    This class is responsible for parsing the front end data store and writing changes to the project files.

    It's a three step process to write changes to the project files:

        1. Re-build the named child to reflect how the information sits in the project files.
        2. Run `update_file_contents` to make changes to project files using a special round trip yaml parser that preserves comments and spacing.
        3. Run `write` to write the changes to the project files.
    """

    def __init__(self, named_children: dict):
        self.named_children = named_children
        self.yaml = ruamel.yaml.YAML(typ="rt")
        self.yaml.indent(mapping=2, sequence=4, offset=2)
        self.yaml.preserve_quotes = True
        self.yaml.width = float("inf")
        self.files_to_write = self.__set_initial_files_to_write_map(named_children)

    def update_file_contents(self):
        for child_name, child_info in self.named_children.items():
            match child_info["status"]:
                case "Unchanged":
                    continue
                case "New":
                    self._new(child_name)
                case "Deleted":
                    self._delete(child_name)
                case "Moved":
                    self._move(child_name)
                case "Modified":
                    self._update(child_name)
                case "Renamed":
                    self._rename(child_name, child_name)

    def write(self):
        """
        Writes changes to all files that have been updated. Setting "typ='rt'" on
        init allows us to keep the original spacing, comments, etc.
        """
        for file, contents in self.files_to_write.items():
            with open(file, "w") as file:
                self.yaml.dump(contents, file)

    def _update(self, child_name: str):
        new_object = self._get_named_child_config(child_name)

        def recurse(current):
            if isinstance(current, dict):
                for key, value in current.items():
                    if isinstance(value, dict) and value.get("name") == child_name:
                        diff_result = diff(value, new_object)
                        apply_diff(value, diff_result)
                        return True
                    elif isinstance(value, (dict, list)):
                        if recurse(value):
                            return True
            elif isinstance(current, list):
                for item in current:
                    if isinstance(item, dict) and item.get("name") == child_name:
                        diff_result = diff(item, new_object)
                        apply_diff(item, diff_result)
                        return True
                    elif isinstance(item, (dict, list)):
                        if recurse(item):
                            return True
            return False

        file_path = self.named_children[child_name]["file_path"]
        # Modify the file contents in place to reflect the changes for the named child
        recurse(self.files_to_write[file_path])

    def _new(self, child_name: str):
        child_dict = self._get_named_child_config(child_name)
        file_path = self.named_children[child_name]["new_file_path"]
        type_key = self.named_children[child_name]["type_key"]

        if self.files_to_write[file_path].get(type_key) is None:
            self.files_to_write[file_path][type_key] = []
        # Inplace append the new child to the type key list
        self.files_to_write[file_path][type_key].append(child_dict)

    def _delete(self, child_name: str, replace_with_reference: bool = False):

        def recurse(current):
            # Handle dictionaries
            if isinstance(current, dict):
                for key, value in list(current.items()):  # Use list to safely iterate
                    if isinstance(value, dict) and value.get("name") == child_name:
                        if replace_with_reference:
                            current[key] = "${ref(" + f"{child_name}" + ")}"
                        else:
                            del current[key]
                        return True
                    elif isinstance(value, (dict, list)):
                        if recurse(value):
                            return True
            # Handle lists
            elif isinstance(current, list):
                for i, item in enumerate(current):
                    if isinstance(item, dict) and item.get("name") == child_name:
                        if replace_with_reference:
                            current[i] = "${ref(" + f"{child_name}" + ")}"
                        else:
                            del current[i]
                        return True
                    elif isinstance(item, (dict, list)):
                        if recurse(item):
                            return True
            return False

        file_path = self.named_children[child_name]["file_path"]
        # Modify the file contents in place to reflect the deletion of the named child
        recurse(self.files_to_write[file_path])

    def _move(self, child_name: str):
        """
        Moving files is just a combination of deleting a file and then creating a new one.
        If the new file path is the same as the old file path, it will move the named child to
        a flat list in the project file.
        """
        # Check if the object is defined inline by looking at its config
        is_inline = self.named_children[child_name].get("is_inline_defined", False)

        self._delete(child_name, replace_with_reference=is_inline)
        self._new(child_name)

    def _rename(self, old_child_name: str, new_child_name: str):
        """Find and replace the old child name with the new child name in all file refs & in the named_child"""
        raise NotImplementedError("Rename is not implemented yet.")

    def _get_named_child_config(self, named_child_name: str) -> dict:
        """
        Returns the reconstructed config for a named child.
        """
        if named_child_name not in self.named_children:
            raise ValueError(
                f"Named child {named_child_name} not found in named_children dictionary."
            )
        return self.__reconstruct_named_child_config(
            self.named_children[named_child_name].get("config")
        )

    def __reconstruct_named_child_config(self, named_child_config: dict) -> dict:
        """ "
        This function will recursivly reconstruct a named child and return a config dictionary with the
        references and inline children returned to the original form found in the file.

        """

        def reconstruct_config(obj) -> dict:
            if isinstance(obj, dict):
                return {k: reconstruct_config(v) for k, v in obj.items()}
            elif isinstance(obj, list):
                return [reconstruct_config(item) for item in obj]
            elif isinstance(obj, str):
                # if the string is a json string with a key of is_inline_defined, return the parsed json
                try:
                    parsed_json = json.loads(obj)
                    if isinstance(parsed_json, dict) and "is_inline_defined" in parsed_json:
                        if parsed_json["is_inline_defined"]:
                            inline_child_object = deepcopy(self.named_children[parsed_json["name"]])
                            inline_config = inline_child_object.get("config")
                            return reconstruct_config(inline_config)
                        else:
                            if parsed_json.get("original_value") is None:
                                raise ValueError(
                                    f"Inline defined named child {parsed_json['name']} does not have an original value set."
                                )
                            else:
                                # return the original reference value so that it matches the original file
                                return parsed_json.get("original_value")
                    else:
                        return obj
                except json.JSONDecodeError:
                    # if the string is not a json string, return the original string
                    return obj
            else:
                # if the object is not a string, return the original object
                return obj

        return reconstruct_config(named_child_config)

    def __set_initial_files_to_write_map(self, named_children: dict) -> dict:

        relevant_files = []
        for value in named_children.values():
            if value.get("status") == "Unchanged":
                continue
            else:
                relevant_files.append(value.get("file_path"))
                relevant_files.append(value.get("new_file_path"))
        relevant_files = [x for x in list(set(relevant_files)) if x is not None]
        files_to_write = {}

        for file_path in relevant_files:
            if file_path is None:
                continue
            elif os.path.isfile(file_path):
                with open(file_path, "r") as file:
                    file_contents = self.yaml.load(file)
                    files_to_write[file_path] = file_contents
            else:
                with open(file_path, "w") as file:
                    file.write("")
                files_to_write[file_path] = {}
        return files_to_write



import click
from visivo.commands.options import project_dir


@click.command()
@project_dir
def init(project_dir):
    """
    Enables a quick set up by writing your source & api credentials to an env file.
    """
    from visivo.logger.logger import Logger

    Logger.instance().debug("Init")

    from visivo.commands.init_phase import init_phase

    init_phase(project_dir)
    Logger.instance().success("Done")



import re
from typing import List
from abc import ABC, abstractmethod
from visivo.models.base.base_model import BaseModel
from visivo.models.base.named_model import NamedModel
from visivo.models.base.project_dag import ProjectDag
from pydantic_core import PydanticCustomError
from pydantic import model_serializer
from visivo.models.dag import (
    all_descendants,
    all_descendants_of_type,
    all_descendants_with_name,
    all_descendants_with_path_match,
    all_nodes_including_named_node_in_graph,
    create_dag_dict,
)
from visivo.models.sources.source import DefaultSource


class ParentModel(ABC):
    @model_serializer(mode="wrap")
    def wrap_serializer(self, serializer, info):
        # Get the default serialized output
        serialized = serializer(self)
        # Check the context to decide whether to include the class type
        if info.context and info.context.get("include_type", False):
            serialized["__type__"] = self.__class__.__name__
        return serialized

    @abstractmethod
    def child_items(self):
        return []

    def dag(self, node_permit_list=None) -> ProjectDag:
        dag = ProjectDag()
        dag.add_node(self)
        self.__build_dag(
            items=self.child_items(),
            parent_item=self,
            dag=dag,
            node_permit_list=node_permit_list,
            root=self,
        )
        self.__dereference_items(
            items=self.child_items(),
            parent_item=self,
            dag=dag,
            node_permit_list=node_permit_list,
            root=self,
        )
        return dag

    def __build_dag(self, items: List, parent_item, dag, node_permit_list, root):
        for item in items:
            if node_permit_list is None or item in node_permit_list:
                dag_item = item
                if (
                    item is None
                    or BaseModel.is_ref(item)
                    or item.__class__.__name__ == "ContextString"
                ):
                    continue
                elif isinstance(item, DefaultSource):
                    name = root.defaults.source_name
                    dag_item = self.__get_dereferenced_item_by_name(
                        name=name,
                        dag=dag,
                        root=root,
                        item=item,
                        parent_item=parent_item,
                    )
                if item.__class__.__name__ == "Test":
                    dag.add_edge(root, dag_item)
                else:
                    dag.add_edge(parent_item, dag_item)
                if isinstance(dag_item, ParentModel):
                    self.__build_dag(
                        items=dag_item.child_items(),
                        parent_item=dag_item,
                        dag=dag,
                        node_permit_list=node_permit_list,
                        root=root,
                    )

    def __dereference_items(self, items: List, parent_item, dag, node_permit_list, root):
        for item in items:
            if node_permit_list is None or item in node_permit_list:
                if item.__class__.__name__ == "ContextString":
                    dereferenced_item = self.__get_dereferenced_item_by_context_string(
                        context_string=item,
                        dag=dag,
                        root=root,
                        parent_item=parent_item,
                    )
                    dag.add_edge(parent_item, dereferenced_item)
                elif BaseModel.is_ref(item):
                    name = NamedModel.get_name(obj=item)
                    dereferenced_item = self.__get_dereferenced_item_by_name(
                        name=name,
                        dag=dag,
                        root=root,
                        item=item,
                        parent_item=parent_item,
                    )
                    dag.add_edge(parent_item, dereferenced_item)
                if isinstance(item, ParentModel):
                    self.__dereference_items(
                        items=item.child_items(),
                        parent_item=item,
                        dag=dag,
                        node_permit_list=node_permit_list,
                        root=root,
                    )

    def descendants(self):
        return all_descendants(dag=self.dag(), from_node=self)

    def descendants_of_type(self, type, dag=None):
        if not dag:
            dag = self.dag()
        return all_descendants_of_type(type=type, dag=dag, from_node=self)

    def nodes_including_named_node_in_graph(self, name):
        return all_nodes_including_named_node_in_graph(name=name, dag=self.dag())

    @staticmethod
    def filtered(pattern, objects) -> List:
        def name_match(obj):
            return re.search(pattern, obj.name)

        return list(filter(name_match, objects))

    def dag_dict(self):
        return create_dag_dict(self.dag())

    def __get_dereferenced_item_by_name(self, name, dag, root, item, parent_item):
        dereferenced_items = all_descendants_with_name(name=name, dag=dag, from_node=root)
        if len(dereferenced_items) == 1:
            return dereferenced_items[0]
        elif len(dereferenced_items) > 1:
            file_paths = set(
                path
                for path in map(lambda ref: ref.file_path, dereferenced_items)
                if path is not None
            )
            raise PydanticCustomError(
                "ambiguous_reference",
                f'The reference "{item}" on item "{parent_item.id()}" points to multiple objects. Check for the duplicated name "{name}" in the following files: {" and ".join(file_paths)}.',
                parent_item.model_dump(),
            )
        else:
            raise PydanticCustomError(
                "bad_reference",
                f'The reference "{item}" on item "{parent_item.id()}" does not point to an object.',
                parent_item.model_dump(),
            )

    def __get_dereferenced_item_by_context_string(self, context_string, dag, root, parent_item):
        if context_string.get_reference():
            return self.__get_dereferenced_item_by_name(
                item=context_string,
                name=context_string.get_reference(),
                dag=dag,
                root=root,
                parent_item=parent_item,
            )
        elif context_string.get_path():
            path = context_string.get_path()
            dereferenced_items = all_descendants_with_path_match(path=path, dag=dag, from_node=root)
            dereferenced_items.sort(key=lambda x: len(x.path), reverse=True)

            if len(dereferenced_items) > 0:
                return dereferenced_items[0]
            else:
                raise PydanticCustomError(
                    "bad_reference",
                    f'The reference "{path}" on item "{parent_item.id()}" does not point to an object.',
                    parent_item.model_dump(),
                )



from queue import Queue
from typing import List
from visivo.models.sources.source import Source
from visivo.jobs.job import Job


class JobTracker:
    def __init__(self):
        self.job_queue = Queue()
        self.enqueued: List[Job] = []
        self.running: List[Job] = []
        self.done: List[Job] = []

    @property
    def all_tracked_job_names(self):
        return (
            set(map(lambda job: job.name, self.enqueued))
            .union(set(map(lambda job: job.name, self.running)))
            .union(set(map(lambda job: job.name, self.done)))
        )

    def track_job(self, job: Job):
        self.__update()
        self.enqueued.append(job)
        self.job_queue.put(job)

    def is_job_name_enqueued(self, job_name: str) -> bool:
        self.__update()
        return job_name in self.all_tracked_job_names

    def is_job_name_done(self, job_name: str) -> bool:
        self.__update()
        return job_name in set(map(lambda job: job.name, self.done))

    def is_job_name_failed(self, job_name: str) -> bool:
        self.__update()
        failed_jobs = list(
            filter(lambda job: job.future and not job.future.result().success, self.done)
        )
        return job_name in set(map(lambda job: job.name, failed_jobs))

    def is_done(self) -> bool:
        self.__update()
        return (len(self.running) + len(self.enqueued)) == 0

    def return_to_queue(self, job: Job):
        self.job_queue.put(job)

    def get_next_job(self) -> Job:
        self.__update()
        return self.job_queue.get(timeout=0.01)

    def empty(self) -> bool:
        return self.job_queue.empty()

    def __update(self):
        self.done = (
            self.done
            + list(filter(lambda job: job.done(), self.running))
            + list(filter(lambda job: job.done(), self.enqueued))
        )
        self.running = list(filter(lambda job: job.running(), self.running)) + list(
            filter(lambda job: job.running(), self.enqueued)
        )

        self.enqueued = list(filter(lambda job: not job.future, self.enqueued))



from flask import Flask
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
from threading import Thread, Event
import time
from visivo.logger.logger import Logger
import os
from flask_socketio import SocketIO
import logging
import socket


class ProjectChangeHandler(FileSystemEventHandler):
    def __init__(self, callback, ignore_patterns=None):
        self.callback = callback
        self.ignore_patterns = ignore_patterns or []
        self.last_event_time = 0
        self.debounce_seconds = 0.5  # Debounce events within 500ms

    def on_modified(self, event):
        if event.is_directory:
            return

        # Only process .yml files
        if not (event.src_path.endswith(".yml") or event.src_path.endswith(".yaml")):
            return

        # Check if file should be ignored
        if any(pattern in event.src_path for pattern in self.ignore_patterns):
            return

        current_time = time.time()
        if current_time - self.last_event_time > self.debounce_seconds:
            Logger.instance().debug(f"Triggering file modified: {event.src_path}")
            self.last_event_time = current_time
            self.callback()


class HotReloadServer:
    @staticmethod
    def find_available_port(start_port=8000, max_attempts=100):
        """Find an available port starting from start_port"""
        for port in range(start_port, start_port + max_attempts):
            try:
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    s.bind(("", port))
                    return port
            except OSError:
                continue
        raise RuntimeError(f"Could not find an available port after {max_attempts} attempts")

    def __init__(self, app: Flask, watch_path: str, ignore_patterns=None):
        self.app = app
        self.watch_path = watch_path
        self.ignore_patterns = ignore_patterns or []
        self.observer = None
        self.server_thread = None
        self.stop_event = Event()
        self.socketio = SocketIO(self.app, cors_allowed_origins="*", async_mode="threading")

        if not os.environ.get("DEBUG"):
            # Suppress Flask logging
            log = logging.getLogger("werkzeug")
            log.setLevel(logging.ERROR)

            # Suppress Flask-SocketIO logging
            log = logging.getLogger("engineio")
            log.setLevel(logging.ERROR)
            log = logging.getLogger("socketio")
            log.setLevel(logging.ERROR)

    def start_file_watcher(self, callback, one_shot=False):
        """Start watching for file changes"""

        def wrapped_callback():
            # Pass one_shot context to the callback
            callback(one_shot=one_shot)
            # Notify clients to refresh after callback completes
            self.socketio.emit("reload")

        event_handler = ProjectChangeHandler(wrapped_callback, self.ignore_patterns)
        self.observer = Observer()
        self.observer.schedule(event_handler, self.watch_path, recursive=True)
        self.observer.start()
        Logger.instance().debug(f"Started file watcher for YML files on {self.watch_path}")

    def run_server(self, host: str, port: int):
        """Run the Flask server in a separate thread"""

        def run():
            self.socketio.run(
                self.app,
                host=host,
                port=port,
                use_reloader=False,
                allow_unsafe_werkzeug=True,
                debug=bool(os.environ.get("DEBUG")),
                log_output=bool(os.environ.get("DEBUG")),
            )

        self.server_thread = Thread(target=run, daemon=True)
        self.server_thread.start()
        Logger.instance().debug(f"Started server on {host}:{port}")

    def serve(
        self,
        host: str,
        port: int,
        on_change_callback=None,
        on_server_ready=None,
        one_shot=False,
    ):
        """Start both the server and file watcher

        Args:
            host: Host to bind to
            port: Port to listen on
            on_change_callback: Callback for file changes
            on_server_ready: Callback to run after server starts
            one_shot: If True, server will shut down after on_server_ready callback completes
        """
        try:
            # Add route for client-side reload script
            @self.app.route("/hot-reload.js")
            def hot_reload_script():
                return """
                    const socket = io();
                    socket.on('reload', () => {
                        console.log('Reloading page...');
                        window.location.reload();
                    });
                """, {
                    "Content-Type": "application/javascript"
                }

            # Start the Flask server
            self.run_server(host, port)

            # Run initialization callback if provided
            if on_server_ready:
                # Pass one_shot context to the callback
                on_server_ready(one_shot=one_shot)
                if one_shot:
                    self.stop()
                    return

            # Start the file watcher if callback provided and not in one_shot mode
            if on_change_callback and not one_shot:
                self.start_file_watcher(on_change_callback, one_shot=one_shot)

            # Keep the main thread alive if not in one_shot mode
            while not one_shot and not self.stop_event.is_set():
                time.sleep(1)

        except KeyboardInterrupt:
            self.stop()

        except Exception as e:
            Logger.instance().error(f"Server error: {str(e)}")
            self.stop()
            raise

    def stop(self):
        """Stop the server and file watcher"""
        if self.observer:
            self.observer.stop()
            self.observer.join()

        self.stop_event.set()

        if self.server_thread:
            # Flask's development server doesn't handle clean shutdowns well
            # We'll let it be terminated when the main process exits
            pass

        Logger.instance().debug("Stopped server and file watcher")



from importlib.resources import files
import json
from jsonschema_rs import ValidationError, validator_for
from pydantic import model_validator
from visivo.models.color_palette import ColorPalette
from visivo.models.trace_props.json_schema_base import JsonSchemaBase, get_message_from_error


class Layout(JsonSchemaBase):

    @model_validator(mode="before")
    @classmethod
    def validate_colorway(cls, data: dict) -> dict:
        v = data.get("colorway", "High Contrast")  # Also handle case where colorway isn't in data

        if v is None:
            return data

        if isinstance(v, str):
            if v not in ColorPalette.PREDEFINED_PALETTES:
                raise ValueError(
                    f"Invalid palette name. Choose from: {', '.join(ColorPalette.PREDEFINED_PALETTES.keys())}"
                )
            data["colorway"] = ColorPalette.PREDEFINED_PALETTES[v]
            return data

        if isinstance(v, list):
            for color in v:
                if not isinstance(color, str):
                    raise ValueError("All colors must be strings")
            return data

        raise ValueError("colorway must be either a palette name or list of colors")

    @model_validator(mode="after")
    def validate_against_schema(self) -> "Layout":
        if not hasattr(Layout, "_schema"):
            schema_path = files("visivo.schema").joinpath(f"layout.schema.json")
            with open(schema_path) as f:
                Layout._schema = json.load(f)

        if not hasattr(Layout, "_validator"):
            Layout._validator = validator_for(Layout._schema)

        validator = Layout._validator
        if not validator:
            raise ValueError("Schema not found for layout")

        try:
            data_dict = self.model_dump()

            validator.validate(data_dict)

        except FileNotFoundError:
            raise ValueError("Schema file not found for layout")
        except json.JSONDecodeError:
            raise ValueError("Invalid JSON in schema file for layout")
        except ValidationError as e:
            message = get_message_from_error(e, Layout._schema)
            raise ValueError(
                f"Validation error for layout at location: {e.instance_path}: {message}"
            )

        return self



import click
import os
from pathlib import Path
from visivo.logger.logger import Logger
from visivo.commands.init_phase import init_phase


def create_phase(project_name=None):
    """Enables a quick set up by writing your source & api credentials to an env file."""
    Logger.instance().success("Creating")
    if not project_name:
        project_name = click.prompt("? Project name", type=str)
    if Path(project_name).exists():
        raise click.ClickException(f"'{project_name}' directory already exists")

    os.makedirs(project_name, exist_ok=True)
    init_phase(project_name)



from visivo.models.dag import all_descendants_of_type
from visivo.models.dashboard import Dashboard
from visivo.models.trace import Trace
import click
import requests
import math
import json
import sys
import asyncio
import aiofiles
import httpx
import os
import contextvars
from time import time
from tenacity import retry, stop_after_attempt, wait_fixed
from visivo.commands.utils import get_profile_file, get_profile_token
from visivo.discovery.discover import Discover
from visivo.logger.logger import Logger
from visivo.parsers.serializer import Serializer
from visivo.parsers.parser_factory import ParserFactory
from visivo.utils import get_dashboards_dir, sanitize_filename

# Limit concurrent uploads to avoid overloading the API
semaphore_3 = asyncio.Semaphore(3)
semaphore_50 = asyncio.Semaphore(50)
attempt = contextvars.ContextVar("attempt")
MAX_ATTEMPTS = 3


@retry(stop=stop_after_attempt(MAX_ATTEMPTS), wait=wait_fixed(2))
async def start_files(file_names, description, form_headers, host, progress):
    """
    Asynchronously uploads trace data files.
    """
    files = list(map(lambda file_name: {"filename": file_name}, file_names))
    url = f"{host}/api/files/direct/start/"
    attempt.set(attempt.get(0) + 1)
    async with semaphore_3:
        try:
            async with httpx.AsyncClient(timeout=30) as client:
                response = await client.post(url, json=files, headers=form_headers)
                response.raise_for_status()
                progress["completed"] += 1
                Logger.instance().success(
                    f"\t{len(file_names)} {description} files created. [{progress['completed']}/{progress['total']}]"
                )
                return response.json()
        except httpx.HTTPStatusError as e:
            Logger.instance().error(
                f"\t[Attempt {attempt.get()}/{MAX_ATTEMPTS}] HTTP error while creating {description} files: {repr(e)} - Response: {e.response.text}"
            )
            raise
        except Exception as e:
            Logger.instance().error(
                f"\t[Attempt {attempt.get()}/{MAX_ATTEMPTS}] Failed to create {description} files: {repr(e)}"
            )
            raise


@retry(stop=stop_after_attempt(MAX_ATTEMPTS), wait=wait_fixed(2))
async def finish_files(file_ids, description, form_headers, host, progress):
    """
    Asynchronously uploads trace data files.
    """
    url = f"{host}/api/files/direct/finish/"
    attempt.set(attempt.get(0) + 1)
    ids = list(map(lambda id: {"id": id}, file_ids))
    async with semaphore_3:
        try:
            async with httpx.AsyncClient(timeout=60) as client:
                response = await client.post(url, json=ids, headers=form_headers)
                response.raise_for_status()
                progress["completed"] += 1
                Logger.instance().success(
                    f"\t{len(file_ids)} {description} files finished [{progress['completed']}/{progress['total']}]"
                )
                return "finished"
        except httpx.HTTPStatusError as e:
            Logger.instance().error(
                f"\t[Attempt {attempt.get()}/{MAX_ATTEMPTS}] HTTP error while finishing files: {repr(e)} - Response: {e.response.text}"
            )
            raise
        except Exception as e:
            Logger.instance().error(
                f"\t[Attempt {attempt.get()}/{MAX_ATTEMPTS}] Failed to finish files: {repr(e)}"
            )
            raise


@retry(stop=stop_after_attempt(MAX_ATTEMPTS), wait=wait_fixed(2))
async def upload_file(name, upload_url, file_name, output_dir, form_headers, progress):
    """
    Asynchronously uploads trace data files.
    """
    attempt.set(attempt.get(0) + 1)
    async with semaphore_50:
        try:
            data_file = f"{output_dir}/{file_name}"
            additional_headers = {}
            if "localhost" in upload_url:
                additional_headers["Content-Disposition"] = f"inline;filename={file_name}"
            async with httpx.AsyncClient(timeout=30) as client:
                async with aiofiles.open(data_file, "rb") as f:
                    response = await client.put(
                        upload_url,
                        content=await f.read(),
                        headers={**form_headers, **additional_headers},
                    )
                    response.raise_for_status()
                    progress["completed"] += 1
                    Logger.instance().success(
                        f"\t'{name}' file uploaded [{progress['completed']}/{progress['total']}]"
                    )
                    return "uploaded"
        except httpx.HTTPStatusError as e:
            Logger.instance().error(
                f"\t[Attempt {attempt.get()}/{MAX_ATTEMPTS}] HTTP error while creating '{name}': {repr(e)} - Response: {e.response.text}"
            )
            raise
        except Exception as e:
            Logger.instance().error(
                f"\t[Attempt {attempt.get()}/{MAX_ATTEMPTS}] Failed to upload data for '{name}': {repr(e)}"
            )
            raise


@retry(stop=stop_after_attempt(MAX_ATTEMPTS), wait=wait_fixed(2))
async def create_trace_records(batch, project_id, json_headers, host, progress):
    """
    Asynchronously creates a trace record on the server.
    """
    body = list(
        map(
            lambda data_file_upload: {
                "name": data_file_upload["name"].split(".")[0],
                "project_id": project_id,
                "data_file_id": data_file_upload["id"],
            },
            batch,
        )
    )
    async with semaphore_3:
        try:
            async with httpx.AsyncClient(timeout=60) as client:
                url = f"{host}/api/traces/"
                response = await client.post(url, json=body, headers=json_headers)
                response.raise_for_status()
                progress["completed"] += 1
                Logger.instance().success(
                    f"\t{len(batch)} traces created [{progress['completed']}/{progress['total']}]"
                )
                return response.json()
        except httpx.HTTPStatusError as e:
            Logger.instance().error(
                f"\t[Attempt {attempt.get()}/{MAX_ATTEMPTS}] HTTP error while creating {len(batch)} traces: {repr(e)} - Response: {e.response.text}"
            )
            raise
        except Exception as e:
            Logger.instance().error(
                f"\t[Attempt {attempt.get()}/{MAX_ATTEMPTS}] Failed to create {len(batch)} traces: {repr(e)}"
            )
            raise


@retry(stop=stop_after_attempt(MAX_ATTEMPTS), wait=wait_fixed(2))
async def create_dashboard_records(
    dashboards, thumbnail_file_uploads, project_id, json_headers, host, progress
):
    """
    Asynchronously creates a trace record on the server.
    """

    def dashboard_body(dashboard):
        thumbnail_file_id = next(
            (
                upload["id"]
                for upload in thumbnail_file_uploads
                if upload["name"].startswith(sanitize_filename(dashboard.name))
            ),
            None,
        )
        dashboard_body = {
            "name": dashboard.name,
            "project_id": project_id,
        }
        if thumbnail_file_id:
            dashboard_body["thumbnail_file_id"] = thumbnail_file_id
        return dashboard_body

    body = list(map(dashboard_body, dashboards))
    async with semaphore_3:
        try:
            async with httpx.AsyncClient(timeout=60) as client:
                url = f"{host}/api/dashboards/"
                response = await client.post(url, json=body, headers=json_headers)
                response.raise_for_status()
                progress["completed"] += 1
                Logger.instance().success(
                    f"\t{len(dashboards)} dashboards created [{progress['completed']}/{progress['total']}]"
                )
                return response.json()
        except httpx.HTTPStatusError as e:
            Logger.instance().error(
                f"\t[Attempt {attempt.get()}/{MAX_ATTEMPTS}] HTTP error while creating {len(dashboards)} dashboards: {repr(e)} - Response: {e.response.text}"
            )
            raise
        except Exception as e:
            Logger.instance().error(
                f"\t[Attempt {attempt.get()}/{MAX_ATTEMPTS}] Failed to create {len(dashboards)} dashboards: {repr(e)}"
            )
            raise


async def process_traces_async(traces, output_dir, project_id, form_headers, json_headers, host):
    """
    Coordinates the asynchronous upload of trace data files and the creation of trace records.
    """
    batch_size = 20
    total_operations = len(traces)  # Each trace has a data upload

    # For each batch upload
    total_operations += 3 * math.ceil(len(traces) / batch_size)
    progress = {"completed": 0, "total": total_operations}

    tasks = []
    for i in range(0, len(traces), batch_size):
        traces_batch = traces[i : i + batch_size]
        traces_batch_file_names = list(map(lambda trace: trace.name, traces_batch))
        create_trace_files_task = start_files(
            traces_batch_file_names, "trace", form_headers, host, progress
        )
        tasks.append(create_trace_files_task)
    data_file_ids = await asyncio.gather(*tasks, return_exceptions=True)
    if any(isinstance(data_file_id, Exception) for data_file_id in data_file_ids):
        raise click.ClickException("Failed to create trace data files.")

    data_file_uploads = [item for sublist in data_file_ids for item in sublist]

    tasks = []
    for data_file_upload in data_file_uploads:
        # Upload data files concurrently
        trace_name = data_file_upload["name"].split(".")[0]
        data_file_task = upload_file(
            trace_name,
            data_file_upload["upload_url"],
            f"traces/{trace_name}/data.json",
            output_dir,
            form_headers,
            progress,
        )
        tasks.append(data_file_task)

    # Wait for all data uploads to complete and gather results
    response_items = await asyncio.gather(*tasks, return_exceptions=True)
    if any(isinstance(item, Exception) for item in response_items):
        raise click.ClickException("Failed to upload trace data.")

    data_file_ids = [item["id"] for item in data_file_uploads]

    tasks = []
    for i in range(0, len(data_file_ids), batch_size):
        batch = data_file_ids[i : i + batch_size]
        create_trace_files_task = finish_files(batch, "trace", form_headers, host, progress)
        tasks.append(create_trace_files_task)

    response_items = await asyncio.gather(*tasks, return_exceptions=True)
    if any(isinstance(item, Exception) for item in response_items):
        raise click.ClickException("Failed to finish trace data files.")

    # Prepare to create trace records based on successful uploads
    tasks = []
    for i in range(0, len(traces), batch_size):
        batch = data_file_uploads[i : i + batch_size]
        task = create_trace_records(batch, project_id, json_headers, host, progress)
        tasks.append(task)

    # Execute the creation of trace records concurrently
    response_items = await asyncio.gather(*tasks, return_exceptions=True)
    if any(isinstance(item, Exception) for item in response_items):
        raise click.ClickException("Failed to create trace records.")


async def process_dashboards_async(
    dashboards, output_dir, project_id, form_headers, json_headers, host
):
    """
    Coordinates the asynchronous upload of thumbnail files.
    """
    dashboards_dir = get_dashboards_dir(output_dir)
    if not os.path.exists(dashboards_dir):
        thumbnail_files = []
    else:
        thumbnail_files = [f for f in os.listdir(dashboards_dir) if f.endswith(".png")]

    total_operations = len(thumbnail_files)  # Each thumbnail has a data upload
    total_operations += 3  # For create, upload and dashboard record creation
    progress = {"completed": 0, "total": total_operations}

    file_names = []
    for dashboard in dashboards:
        sanitized_name = sanitize_filename(dashboard.name)
        if os.path.exists(f"{dashboards_dir}/{sanitized_name}.png"):
            file_names.append(f"{sanitized_name}.png")

    create_thumbnail_files_task = start_files(file_names, "thumbnail", form_headers, host, progress)

    thumbnail_file_uploads_nested = await asyncio.gather(
        create_thumbnail_files_task, return_exceptions=True
    )
    if any(isinstance(data_file_id, Exception) for data_file_id in thumbnail_file_uploads_nested):
        raise click.ClickException("Failed to create thumbnail files.")

    thumbnail_file_uploads = [item for sublist in thumbnail_file_uploads_nested for item in sublist]

    tasks = []
    for thumbnail_file_upload in thumbnail_file_uploads:
        # Upload data files concurrently
        file_name = thumbnail_file_upload["name"]
        dashboard = next(
            (
                dashboard
                for dashboard in dashboards
                if sanitize_filename(dashboard.name) == file_name.split(".")[0]
            ),
            None,
        )
        data_file_task = upload_file(
            dashboard.name,
            thumbnail_file_upload["upload_url"],
            f"dashboards/{file_name}",
            output_dir,
            form_headers,
            progress,
        )
        tasks.append(data_file_task)

    await asyncio.gather(*tasks)
    uploaded_ids = [item["id"] for item in thumbnail_file_uploads]

    # Finish the upload process
    tasks = []
    tasks.append(finish_files(uploaded_ids, "thumbnail", form_headers, host, progress))
    response_items = await asyncio.gather(*tasks, return_exceptions=True)
    if any(isinstance(item, Exception) for item in response_items):
        raise click.ClickException("Failed to finish thumbnail files.")

    tasks = []
    tasks.append(
        create_dashboard_records(
            dashboards, thumbnail_file_uploads, project_id, json_headers, host, progress
        )
    )

    # Execute the creation of trace records concurrently
    response_items = await asyncio.gather(*tasks, return_exceptions=True)
    if any(isinstance(item, Exception) for item in response_items):
        raise click.ClickException("Failed to create dashboard records.")


def deploy_phase(working_dir, user_dir, output_dir, stage, host):
    """
    Synchronous function to manage the deployment, including initiating asynchronous operations.
    """
    deploy_start_time = time()
    # Retrieve profile token for authentication
    Logger.instance().debug("Retrieving profile token...")
    profile_file = get_profile_file(home_dir=user_dir)
    profile_token = get_profile_token(profile_file)
    Logger.instance().info(f"Found Profile token: {profile_file}")

    # Discover and parse project details
    Logger.instance().info("")
    Logger.instance().debug("Compiling project details...")
    discover = Discover(working_dir=working_dir, home_dir=user_dir, output_dir=output_dir)
    parser = ParserFactory().build(project_file=discover.project_file, files=discover.files)
    project = parser.parse()
    serializer = Serializer(project=project)
    project_json = json.loads(serializer.dereference().model_dump_json(exclude_none=True))
    Logger.instance().success(f"Project Compiled in {time() - deploy_start_time:.2f} seconds")

    # Prepare request payloads and headers
    body = {
        "project_json": project_json,
        "name": project_json["name"],
        "cli_version": project_json["cli_version"],
        "stage": stage,
    }
    json_headers = {
        "content-type": "application/json",
        "Authorization": f"Api-Key {profile_token}",
    }
    form_headers = {
        "Authorization": f"Api-Key {profile_token}",
    }

    # Upload the project information (synchronous)
    Logger.instance().info("Uploading project information...")
    upload_project_start_time = time()
    url = f"{host}/api/projects/"
    response = requests.post(url, data=json.dumps(body), headers=json_headers)
    if response.status_code == 401:
        raise click.ClickException(f"Token not authorized for host: {host}")
    if response.status_code == 404:
        raise click.ClickException(f"404 error raised. Does your user have an account?")
    if response.status_code == 201:
        Logger.instance().success(
            f"Project uploaded in {time() - upload_project_start_time:.2f} seconds"
        )
        project_data = response.json()
        project_id = project_data["id"]
        project_url = project_data["url"]

        # Process traces
        Logger.instance().info(f"")
        Logger.instance().info("Processing trace uploads and record creations...")
        process_traces_start_time = time()

        traces = []
        dag = project.dag()
        dashboards = all_descendants_of_type(type=Dashboard, dag=dag)
        for dashboard in dashboards:
            traces.extend(all_descendants_of_type(type=Trace, dag=dag, from_node=dashboard))
        traces = list(set(traces))
        asyncio.run(
            process_traces_async(
                traces=traces,
                output_dir=output_dir,
                project_id=project_id,
                form_headers=form_headers,
                json_headers=json_headers,
                host=host,
            )
        )
        Logger.instance().info(
            f"Trace uploads and record creations completed in {time() - process_traces_start_time:.2f} seconds"
        )

        # Process thumbnails
        Logger.instance().info("Processing dashboard uploads...")
        process_thumbnails_start_time = time()
        dashboards = project.descendants_of_type(type=Dashboard)

        asyncio.run(
            process_dashboards_async(
                dashboards=dashboards,
                output_dir=output_dir,
                project_id=project_id,
                form_headers=form_headers,
                json_headers=json_headers,
                host=host,
            )
        )
        Logger.instance().info(
            f"Thumbnail uploads completed in {time() - process_thumbnails_start_time:.2f} seconds"
        )

        # Deploy the project
        url = f"{host}/api/projects/{project_id}/"
        response = requests.put(
            url, data=json.dumps({"deploy_finished_at": "now"}), headers=json_headers
        )
        if response.status_code == 200:
            Logger.instance().success(
                f"Deployment completed in {time() - deploy_start_time:.2f} seconds"
            )
        else:
            Logger.instance().info(f"Deployment failed in {time() - deploy_start_time:.2f} seconds")
            sys.exit(1)
        return project_url
    else:
        Logger.instance().info(f"Deployment failed in {time() - deploy_start_time:.2f} seconds")
        raise click.ClickException(f"There was an unexpected error: {response.content}")



from typing import List

import pydantic
from visivo.models.models.model import Model, TableModelName
from pydantic import Field
from visivo.models.sources.duckdb_source import DuckdbSource
import io
import click
import os
import polars as pl
import csv


class CsvScriptModel(Model):
    """
    CSV Script Models are a type of model that executes a command with a given set of args.
    This command needs to return a well formatted :fontawesome-solid-file-csv: with a header row to stdout.

    Visivo will be able to access the generate file as a model by storing a duckdb file in the source directory.

    !!! example {% raw %}

        === "Echo"

            Echoing all of your data is probably not a very practical example, but it does nicely demonstrate how the feature works!
            ``` yaml
            models:
              - name: csv
                table_name: csv
                args:
                    - echo
                    - |
                      x,y
                      1,9
                      2,1
                      3,2
                      4,3
                      5,5
                      6,8
            ```

        === "Python Script"

            In this example we'll use python to generate a csv of processes running on your machine and make that csv available to Visivo as
            a model for analysis.
            ``` python title="created_processes_csv.py"
            import subprocess
            import csv
            import sys

            # Define the CSV file to write
            csv_file = "data/processes.csv"

            # Execute the 'ps aux' command
            result = subprocess.run(["ps", "aux"], stdout=subprocess.PIPE, text=True)

            # Split the output into lines
            lines = result.stdout.strip().split("/n")

            # Write CSV to stdout
            writer = csv.writer(sys.stdout)
            writer.writerow(
                ["USER","PID","%CPU","%MEM","VSZ","RSS","TTY","STAT","START","TIME","COMMAND"]
            )  # Header

            for line in lines[1:]:  # Skip the header line from the ps output
                row = line.split(None, 10)  # Split on whitespace, but only for the first 10 columns
                writer.writerow(row)
            ```
            With your script ready to go, all you have to do is convert `python create_processes_csv.py` into the args list format in a model.
            ``` yaml
            models:
              - name: processes
                table_name: processes
                args:
                  - python
                  - create_processes_csv.py
            ```{% endraw %}

        === "CSV File"

            One of the best use cases for this type of model is to store a static csv in your project and cat it into a model.
            This great because it's simple and allows you to version control your csv data.
            ``` csv title="file.csv"
            columns,go,up,here
            1,text,more text,6
            2,stuff,more stuff,7
            ```
            Then just `cat` the csv file in a model.
            ``` yaml
            models:
              - name: file_model
                table_name: file_model
                args:
                  - cat
                  - file.csv
            ```

    The args are python subprocess list args and you can read their source [documentation here](https://docs.python.org/3/library/subprocess.html#subprocess.CompletedProcess.args).
    """

    name: str = pydantic.Field(
        ..., description="The unique name of the object across the entire project."
    )

    table_name: TableModelName = Field(
        "model", description="The name to give the resulting models table"
    )

    args: List[str] = Field(description="An array of the variables that build your command to run.")

    allow_empty: bool = Field(
        default=False, description="Whether to allow the command to return an empty csv."
    )

    @property
    def sql(self):
        return f"select * from {self.table_name}"

    def get_duckdb_source(self, output_dir) -> DuckdbSource:
        os.makedirs(f"{output_dir}/models", exist_ok=True)
        return DuckdbSource(
            name=f"model_{self.name}_generated_source",
            database=f"{output_dir}/models/{self.name}.duckdb",
            type="duckdb",
        )

    def validate_stream_is_csv(self, stream: io.StringIO):
        try:
            # First check if the stream is empty
            content = stream.read()
            stream.seek(0)  # Reset stream position

            if not content.strip():
                if self.allow_empty:
                    return
                raise click.ClickException(
                    f"The command {self.name} did not return any data. Verify command's output and try again."
                )

            # Now try to parse the CSV
            reader = csv.reader(io.StringIO(content))
            rows = list(reader)

            if not rows:
                if self.allow_empty:
                    return
                raise click.ClickException(
                    f"The command {self.name} did not return any data. Verify command's output and try again."
                )

            expected = len(rows[0])
            for number, row in enumerate(rows, start=1):
                if len(row) == 0:
                    continue
                if len(row) != expected:
                    raise click.ClickException(
                        f"CSV parsing error for node:{self.name} model's command. Row {number} has {len(row)} columns but expected {expected}. Verify command's output and try again."
                    )
        except csv.Error as e:
            if os.environ.get("STACKTRACE"):
                raise e
            raise click.ClickException(
                f"CSV parsing error for {self.name} model's command. Verify command's output is a valid csv and try again. Full error: {repr(e)}"
            )
        finally:
            stream.seek(0)

    def insert_csv_to_duckdb(self, output_dir):
        import subprocess

        try:
            process = subprocess.Popen(self.args, stdout=subprocess.PIPE)
            source = self.get_duckdb_source(output_dir)
            with source.connect() as connection:
                csv = io.StringIO(process.stdout.read().decode())

                self.validate_stream_is_csv(csv)

                data_frame = pl.read_csv(csv)

                # Register the Polars DataFrame as a DuckDB view
                duckdb_conn = connection
                duckdb_conn.register("data_frame", data_frame)
                duckdb_conn.execute(
                    f"CREATE TABLE IF NOT EXISTS {self.table_name} AS SELECT * FROM data_frame"
                )
                duckdb_conn.execute(f"DELETE FROM {self.table_name}")
                duckdb_conn.execute(f"INSERT INTO {self.table_name} SELECT * FROM data_frame")
        except Exception as e:
            raise click.ClickException(
                f"Error parsing or generating the csv output of node:{self.name} model's command. Verify command's output and try again. Full error: {repr(e)}"
            )



import hashlib
import yaml
import json
import os
from pathlib import Path
import re
import click
from visivo.logger.logger import Logger
from visivo.models.base.query_string import QueryString
from visivo.templates.render_yaml import render_yaml
from visivo.parsers.yaml_ordered_dict import YamlOrderedDict
import importlib.resources as resources

VIEWER_PATH = resources.files("visivo") / "viewers" / "local"
DIST_PATH = resources.files("visivo") / "viewers" / "dist"

SCHEMA_FILE = resources.files("visivo") / "src" / "visivo_project_schema.json"

PROJECT_CHILDREN = [
    "alerts",
    "selectors",
    "sources",
    "models",
    "traces",
    "tables",
    "charts",
    "dashboards",
    "destinations",
]


def get_dashboards_dir(output_dir):
    return os.path.join(output_dir, "dashboards")


def sanitize_filename(name):
    """Generate a hash of the name for safe filenames"""
    return hashlib.md5(name.encode()).hexdigest()


def yml_to_dict(relative_path):
    with open(relative_path, "r") as file:
        yaml_dict = yaml.safe_load(Path(relative_path).read_text())
        return dict(yaml_dict)


def list_all_ymls_in_dir(path):
    dir_path = Path(path)
    file_paths = list(dir_path.rglob("*.yml")) + list(dir_path.rglob("*.yaml"))
    file_paths = [p for p in file_paths if os.path.isfile(p)]
    return file_paths


def json_to_dict(relative_path):
    with open(relative_path) as json_file:
        data = json.load(json_file)
    return dict(data)


def file_from_path(path):
    return str(path).split("/")[-1].split(".")[0]


def remove_file_from_path(path):
    return "/".join(str(path).split("/")[:-1])


def sql_to_string(path):
    with open(path, "r") as file:
        sqlFile = file.read()
    return sqlFile.strip().lower()


def extract_table_name_from_create(sql_string):
    create_statement = sql_string.replace("\n", " ").split(" as ")[0].strip()
    if "create" in create_statement:
        return create_statement.split(" ")[-1]
    else:
        return None


def listify(string_or_list):
    if isinstance(string_or_list, str):
        return [string_or_list]
    elif isinstance(string_or_list, list):
        return string_or_list
    elif string_or_list == None:
        return [None]
    else:
        raise ValueError("Passed object that was neither a string, list or None")


def error_if_true(bool: bool, message: str):
    if bool:
        raise Exception(message)


def extract_value_from_function(function_value, function_name):
    if function_value is None:
        return None
    if isinstance(function_value, QueryString):
        return function_value.get_value()
    query_string = QueryString(str(function_value))
    if query_string.get_value():
        return query_string.get_value()
    pattern = r"{}\((.+)\)".format(re.escape(function_name))
    match = re.search(pattern, str(function_value))
    if not match:
        return None
    value = match.group(1)
    return value.strip()


def set_location_recursive_items(dictionary, file):
    if isinstance(dictionary, YamlOrderedDict):
        for key, value in dictionary._key_locs.items():
            dictionary._key_locs[key] = value.replace("<unicode string>", file)

        for key, value in dictionary._value_locs.items():
            dictionary._value_locs[key] = value.replace("<unicode string>", file)

        for key, value in dictionary.items():
            if isinstance(value, dict):
                set_location_recursive_items(value, file)
            if isinstance(value, list):
                for item in value:
                    set_location_recursive_items(item, file)


def load_yaml_file(file):
    with open(file, "r") as stream:
        template_string = stream.read()
        try:
            loaded = yaml.safe_load(render_yaml(template_string))
            set_location_recursive_items(loaded, str(file))
            return loaded
        except yaml.YAMLError as exc:
            if hasattr(exc, "problem_mark"):
                mark = exc.problem_mark
                error_location = f"Invalid yaml in project\n  Location: {str(file)}:{mark.line + 1}[{mark.column + 1}]\n  Issue: {exc.problem}"
                raise click.ClickException(error_location)
            else:
                raise click.ClickException(exc)


def nested_dict_from_dotted_keys(flat_dict):
    nested_dict = {}
    for key, value in flat_dict.items():
        parts = key.split(".")
        d = nested_dict
        for part in parts[:-1]:
            if part not in d:
                d[part] = {}
            d = d[part]
        if isinstance(value, dict):
            d[parts[-1]] = nested_dict_from_dotted_keys(value)
        else:
            d[parts[-1]] = value
    return nested_dict


def combine_dict_properties(input_dict):
    combined = {}
    for outer_key, inner_dict in input_dict.items():
        for key, value in inner_dict.items():
            if key not in combined:
                combined[key] = []
            if isinstance(value, list):
                combined[key].extend(value)
            else:
                combined[key].append(value)
    return combined


def merge_dicts(dict1, dict2):
    merged = dict1.copy()
    for key, value in dict2.items():
        if isinstance(value, dict) and key in merged and isinstance(merged[key], dict):
            merged[key] = merge_dicts(merged[key], value)
        else:
            merged[key] = value
    return merged



from visivo.server.models.base import (
    Base,
    Column,
    String,
    Integer,
    DateTime,
    ForeignKey,
    relationship,
)
from datetime import datetime


class ResultModel(Base):
    __tablename__ = "results"

    id = Column(Integer, primary_key=True)
    worksheet_id = Column(String, ForeignKey("worksheets.id", ondelete="CASCADE"))
    results_json = Column(String)
    query_stats_json = Column(String)
    created_at = Column(DateTime, default=datetime.utcnow)

    worksheet = relationship("WorksheetModel", back_populates="results")

    def to_dict(self):
        return {
            "results_json": self.results_json,
            "query_stats_json": self.query_stats_json,
            "created_at": self.created_at.isoformat() if self.created_at else None,
        }



from deepmerge import always_merger
from typing import Dict, List
from pathlib import Path
from pydantic import ValidationError
from visivo.parsers.line_validation_error import LineValidationError
from visivo.parsers.yaml_ordered_dict import setup_yaml_ordered_dict
from visivo.utils import load_yaml_file, PROJECT_CHILDREN
from visivo.models.project import Project


class CoreParser:
    def __init__(self, project_file: Path, files: List[Path]):
        self.files = files
        self.project_file = project_file
        setup_yaml_ordered_dict()

    def parse(self) -> Project:
        return self.__build_project()

    def merge_data_files(self):
        return self.__merged_project_data()

    def project_file_data(self):
        return load_yaml_file(self.project_file)

    def __build_project(self):
        data = self.__merged_project_data()
        data["project_file_path"] = str(self.project_file)
        try:
            project = Project(**data)
        except ValidationError as validation_error:
            raise LineValidationError(validation_error=validation_error, files=self.files)
        return project

    def __merged_project_data(self):
        project_data = self.project_file_data()

        data_files = {}
        for file in self.files:
            if file == self.project_file:
                continue
            data_files[file] = load_yaml_file(file)

        return self.__merge_data_into_project(project_data=project_data, data_files=data_files)

    def __merge_data_into_project(self, project_data: dict, data_files: dict):
        for index, (file_path, data_file) in enumerate(data_files.items()):
            for key_to_merge in PROJECT_CHILDREN.copy():
                if key_to_merge in data_file:
                    base_merge = []
                    if key_to_merge in project_data:
                        base_merge = project_data[key_to_merge]
                        if index == 0:
                            self.__recursively_add_file_path(base_merge, self.project_file)
                    file_to_merge = data_file[key_to_merge]
                    self.__recursively_add_file_path(file_to_merge, file_path)
                    project_data[key_to_merge] = always_merger.merge(base_merge, file_to_merge)
        return project_data

    def __recursively_add_file_path(self, obj, file_path: str):
        """
        Modifies the input object by adding the file path to the named model objects.
        """
        if isinstance(obj, dict):
            if obj.get("name") is not None:
                obj["file_path"] = str(file_path)
            for value in obj.values():
                self.__recursively_add_file_path(value, file_path)
        elif isinstance(obj, list):
            for item in obj:
                self.__recursively_add_file_path(item, file_path)
        else:
            pass



from enum import Enum
from typing import List, Optional
from pydantic import Field, field_serializer, model_validator

from visivo.models.alert import Alert
from visivo.models.base.base_model import BaseModel
from visivo.models.base.eval_string import EvalString
from visivo.models.base.named_model import NamedModel
from visivo.models.base.parent_model import ParentModel

"""
Tests allow you to assert on the computed values that are the output of a trace.  The tests are run with the `visivo test` command.

### Example
``` yaml
tests:
  - name: Test One
    if: ${ ref(Tested Trace).props.type } == "scatter"
    assertions:
      - >{ sum( ${ ref(Tested Trace).props.x) } ) == 7 }
      - >{ ${ ref(Tested Trace).props.x[0] } == 1 } 
traces:
    - name: Tested Trace
      model: ref(model)
      columns:
          account_name: account_name
      props:
          type: scatter
          x: column(project_created_at)
          y: column(project_name)
      tests:
        - assertions: 
            - >{ sum( ${ ref(Tested Trace).props.x) } ) == 7 }
            - >{ ${ ref(Tested Trace).props.x[0] } == 1 } 
          alerts:
            - ${ ref(Alert One) }
```
The [numpy](https://numpy.org/doc/stable/index.html) libraries are available for testing.
"""


class OnFailureEnum(str, Enum):
    exit = "exit"
    continue_ = "continue"


class Test(NamedModel, ParentModel):
    if_: Optional[EvalString] = Field(None, alias="if")
    on_failure: OnFailureEnum = Field(OnFailureEnum.exit)
    assertions: List[EvalString] = Field(None)

    @model_validator(mode="before")
    def rename_if(cls, values):
        if "if_" in values:
            values["if"] = values["if_"]
            del values["if_"]
        return values

    __test__ = False

    def child_items(self):
        assertion_contexts = set()
        for assertion in self.assertions:
            assertion_contexts.update(assertion.get_context_strings())
        if self.if_:
            assertion_contexts.update(self.if_.get_context_strings())
        return list(assertion_contexts)



from typing import Any, Union
import json
from typing_extensions import Annotated
from pydantic import Discriminator, Tag
from visivo.models.dashboard import Dashboard
from visivo.models.dashboards.external_dashboard import ExternalDashboard
from visivo.logger.logger import Logger
from visivo.parsers.yaml_ordered_dict import YamlOrderedDict


def get_dashboard_discriminator_value(value: Any) -> str:
    if isinstance(value, (dict, YamlOrderedDict)):
        if "href" in value:
            return "external"
        elif "rows" in value:
            return "internal"
    elif hasattr(value, "href"):
        return "external"
    elif hasattr(value, "rows"):
        return "internal"
    return "internal"


DashboardField = Annotated[
    Union[
        Annotated[Dashboard, Tag("internal")],
        Annotated[ExternalDashboard, Tag("external")],
    ],
    Discriminator(get_dashboard_discriminator_value),
]



from abc import ABC, abstractmethod
from typing import Any, Optional
import click
from pydantic import PrivateAttr
from visivo.models.sources.source import Source
from sqlalchemy import create_engine, event, text
from sqlalchemy.pool import NullPool
from visivo.logger.logger import Logger
import polars as pl
from copy import deepcopy
import pyarrow as pa


class SqlalchemySource(Source, ABC):

    _engine: Any = PrivateAttr(default=None)
    after_connect: Optional[str] = None

    @abstractmethod
    def get_dialect(self):
        raise NotImplementedError(f"No dialect method implemented for {self.type}")

    def read_sql(self, query: str):
        with self.connect() as connection:
            query = text(query)
            results = connection.execute(query)
            columns = list(results.keys())
            data = results.fetchall()
            results.close()
        # Convert to dict of columns for Polars
        if data:
            data_dict = {col: [row[i] for row in data] for i, col in enumerate(columns)}
            return pl.DataFrame(data_dict)
        else:
            # No data, just return empty DataFrame with columns
            schema = {col: pl.String for col in columns}
            return pl.DataFrame({col: [] for col in columns}, schema=schema)

    def get_connection(self):

        try:
            connection = (
                self.get_engine().connect()
            )  # I wonder if creating mutltiple engines is part of the problem.
            if hasattr(self, "attach") and self.attach:
                for attachment in self.attach:
                    connection.execute(
                        text(
                            f"attach database '{attachment.source.database}' as {attachment.schema_name};"
                        )
                    )
            return connection
        except Exception as err:
            raise click.ClickException(
                f"Error connecting to source '{self.name}'. Ensure the database is running and the connection properties are correct. Full Error: {str(err)}"
            )

    def get_engine(self):

        if not self._engine:

            Logger.instance().debug(f"Creating engine for Source: {self.name}")
            self._engine = create_engine(
                self.url(), poolclass=NullPool, connect_args=self.connect_args()
            )

            @event.listens_for(self._engine, "connect")
            def connect(dbapi_connection, connection_record):
                if self.after_connect:
                    cursor_obj = dbapi_connection.cursor()
                    cursor_obj.execute(self.after_connect)
                    cursor_obj.close()

        return self._engine

    def __deepcopy__(self, memo):
        copied = self.model_copy(deep=False)
        # manually deepcopy only safe attrs
        for name, val in self.__dict__.items():
            if name != "_plugin_module":
                setattr(copied, name, deepcopy(val, memo))
        return copied

    def connect_args(self):
        return {}



from visivo.logger.logger import Logger
from visivo.models.models.csv_script_model import CsvScriptModel
from visivo.jobs.job import (
    Job,
    JobResult,
    format_message_failure,
    format_message_success,
    start_message,
)
from time import time


def action(csv_script_model: CsvScriptModel, output_dir):
    Logger.instance().info(start_message("CsvScriptModel", csv_script_model))
    try:
        start_time = time()
        csv_script_model.insert_csv_to_duckdb(output_dir=output_dir)
        success_message = format_message_success(
            details=f"Updated data for model \033[4m{csv_script_model.name}\033[0m",
            start_time=start_time,
            full_path=csv_script_model.get_duckdb_source(output_dir=output_dir).database,
        )
        return JobResult(item=csv_script_model, success=True, message=success_message)
    except Exception as e:
        failure_message = format_message_failure(
            details=f"Failed query for model \033[4m{csv_script_model.name}\033[0m",
            start_time=start_time,
            full_path=csv_script_model.get_duckdb_source(output_dir=output_dir).database,
            error_msg=str(repr(e)),
        )
        return JobResult(item=csv_script_model, success=False, message=failure_message)


def job(csv_script_model, output_dir: str):
    return Job(
        item=csv_script_model,
        source=csv_script_model.get_duckdb_source(output_dir),
        action=action,
        csv_script_model=csv_script_model,
        output_dir=output_dir,
    )



import re
import os
import json
from flask import Flask, send_from_directory, request, jsonify, Response, send_file
import datetime
from visivo.models.project import Project
from visivo.parsers.serializer import Serializer
from visivo.utils import VIEWER_PATH, SCHEMA_FILE
from visivo.logger.logger import Logger
from visivo.server.project_writer import ProjectWriter
from visivo.server.repositories.worksheet_repository import WorksheetRepository
from visivo.server.text_editors import get_editor_configs
import subprocess
import hashlib


class FlaskApp:
    def __init__(self, output_dir, project: Project):
        self.app = Flask(__name__, static_folder=output_dir, static_url_path="/data")

        self._project_json = (
            Serializer(project=project).dereference().model_dump_json(exclude_none=True)
        )
        self._project = project

        self.app.config["SEND_FILE_MAX_AGE_DEFAULT"] = 0
        self.worksheet_repo = WorksheetRepository(os.path.join(output_dir, "worksheets.db"))

        @self.app.route("/data/<trace_name>/data.json")
        def serve_trace_data(trace_name):
            try:
                trace_dir = os.path.join(output_dir, "traces", trace_name)
                if not os.path.exists(trace_dir):
                    return (
                        jsonify({"message": f"Trace directory not found: {trace_name}"}),
                        404,
                    )
                return send_from_directory(trace_dir, "data.json")
            except Exception as e:
                Logger.instance().error(f"Error serving trace data: {str(e)}")
                return jsonify({"message": str(e)}), 500

        @self.app.route("/data/explorer.json")
        def explorer():
            if os.path.exists(f"{output_dir}/explorer.json"):
                with open(f"{output_dir}/explorer.json", "r") as f:
                    return json.load(f)
            else:
                return json.dumps({})

        @self.app.route("/data/schema.json")
        def schema():
            if os.path.exists(SCHEMA_FILE):
                return send_file(SCHEMA_FILE)
            else:
                return (
                    jsonify({"message": f"Schema file not found: {SCHEMA_FILE}"}),
                    404,
                )

        @self.app.route("/api/project/named_children", methods=["GET"])
        def named_children():
            named_children = self._project.named_child_nodes()
            if named_children:
                return jsonify(named_children)
            else:
                return jsonify({})

        @self.app.route("/api/project/project_file_path", methods=["GET"])
        def project_file_path():
            project_file_path = self._project.project_file_path
            if project_file_path:
                return jsonify(project_file_path)
            else:
                return jsonify({})

        @self.app.route("/api/project/write_changes", methods=["POST"])
        def write_changes():

            data = request.get_json()
            if not data:
                return jsonify({"message": "No data provided"}), 400

            try:
                project_writer = ProjectWriter(data)
                project_writer.update_file_contents()
                project_writer.write()
                return jsonify({"message": "Changes written successfully"}), 200
            except Exception as e:
                Logger.instance().error(f"Error writing changes: {str(e)}")
                return jsonify({"message": str(e)}), 500

        @self.app.route("/api/query/<project_id>", methods=["POST"])
        def execute_query(project_id):
            try:
                data = request.get_json()
                if not data or "query" not in data:
                    return jsonify({"message": "No query provided"}), 400

                query = data["query"]
                source_name = data.get("source")
                Logger.instance().info(f"Executing query with source: {source_name}")
                worksheet_id = data.get("worksheet_id")  # New: Get worksheet_id if provided

                # Get the appropriate source based on the request
                source = None
                if source_name:
                    source = next(
                        (s for s in self._project.sources if s.name == source_name),
                        None,
                    )

                if not source and self._project.defaults and self._project.defaults.source_name:
                    source = next(
                        (
                            s
                            for s in self._project.sources
                            if s.name == self._project.defaults.source_name
                        ),
                        None,
                    )

                if not source and self._project.sources:
                    source = self._project.sources[0]

                if not source:
                    return jsonify({"message": "No source configured"}), 400

                Logger.instance().info(f"Executing query with source: {source.name}")

                # Execute the query using read_sql
                result = source.read_sql(query)

                # Transform the result into the expected format
                if result is None or result.height == 0:
                    response_data = {"columns": [], "rows": []}
                else:
                    response_data = {
                        "columns": list(result.columns),
                        "rows": result.to_dicts(),
                    }

                # If worksheet_id is provided, save the results
                if worksheet_id:
                    query_stats = {
                        "timestamp": datetime.datetime.utcnow().isoformat(),
                        "source": source.name,
                    }
                    self.worksheet_repo.save_results(
                        worksheet_id, json.dumps(response_data), json.dumps(query_stats)
                    )

                return jsonify(response_data), 200

            except Exception as e:
                Logger.instance().error(f"Query execution error: {str(e)}")
                return jsonify({"message": str(e)}), 500

        @self.app.route("/api/trace/<trace_name>/query", methods=["GET"])
        def get_trace_query(trace_name):
            try:
                query_file_path = f"{output_dir}/traces/{trace_name}/query.sql"
                if not os.path.exists(query_file_path):
                    return (
                        jsonify({"message": f"Query file not found for trace: {trace_name}"}),
                        404,
                    )

                with open(query_file_path, "r") as f:
                    query_contents = f.read()

                return jsonify({"query": query_contents}), 200
            except Exception as e:
                Logger.instance().error(f"Error reading trace query: {str(e)}")
                return jsonify({"message": str(e)}), 500

        @self.app.route("/data/error.json")
        def error():
            if os.path.exists(f"{output_dir}/error.json"):
                with open(f"{output_dir}/error.json", "r") as error_file:
                    return error_file.read()
            else:
                return json.dumps({})

        @self.app.route("/data/project.json")
        def projects():
            return {
                "id": "id",
                "project_json": json.loads(self._project_json),
                "created_at": datetime.datetime.now().isoformat(),
            }

        @self.app.route("/data/project_history.json")
        def project_history():
            return [
                {
                    "id": "id",
                    "created_at": datetime.datetime.now().isoformat(),
                }
            ]

        @self.app.route("/", defaults={"path": "index.html"})
        @self.app.route("/<path:path>")
        def viewer_file(path):
            regex = r"\S*(\.png|\.ico|\.js|\.css|\.webmanifest|\.js\.map|\.css\.map)$"
            if re.match(regex, path):
                return send_from_directory(VIEWER_PATH, path)

            # For HTML responses, read the file and inject our scripts
            with open(os.path.join(VIEWER_PATH, "index.html"), "r") as f:
                html = f.read()

            # Add socket.io client and our hot reload script
            scripts = """
                <script src="https://cdnjs.cloudflare.com/ajax/libs/socket.io/4.0.1/socket.io.js"></script>
                <script src="/hot-reload.js"></script>
            """
            html = html.replace("</head>", f"{scripts}</head>")

            return html

        @self.app.route("/data/dashboards/<dashboard_name>", methods=["GET"])
        def get_dashboard(dashboard_name):
            dashboard_name_hash = hashlib.md5(dashboard_name.encode()).hexdigest()
            thumbnail_path = os.path.join("dashboards", f"{dashboard_name_hash}.png")
            exists = os.path.exists(os.path.join(output_dir, thumbnail_path))

            return {
                "id": dashboard_name,
                "name": dashboard_name,
                "signed_thumbnail_file_url": (
                    f"/data/dashboards/{dashboard_name_hash}.png" if exists else None
                ),
            }

        @self.app.route("/data/dashboards/<dashboard_name_hash>.png", methods=["GET"])
        def get_thumbnail(dashboard_name_hash):
            try:
                thumbnail_path = os.path.join("dashboards", f"{dashboard_name_hash}.png")
                if not os.path.exists(os.path.join(output_dir, thumbnail_path)):
                    Logger.instance().debug(f"Thumbnail not found at path: {thumbnail_path}")
                    return Response(status=404)

                return send_from_directory(output_dir, thumbnail_path)
            except Exception as e:
                Logger.instance().error(f"Error retrieving thumbnail: {str(e)}")
                return jsonify({"message": str(e)}), 500

        @self.app.route("/data/dashboards/<dashboard_name_hash>.png", methods=["POST"])
        def create_thumbnail(dashboard_name_hash):
            try:
                if "file" not in request.files:
                    return jsonify({"message": "No file provided"}), 400

                file = request.files["file"]
                if file.filename == "" or not file.filename.endswith(".png"):
                    return jsonify({"message": "Invalid file - must be a PNG"}), 400

                dashboard_dir = os.path.join(output_dir, "dashboards")
                os.makedirs(dashboard_dir, exist_ok=True)

                thumbnail_path = os.path.join(dashboard_dir, f"{dashboard_name_hash}.png")

                file.save(thumbnail_path)

                return jsonify(
                    {
                        "signed_thumbnail_file_url": f"/data/dashboards/{dashboard_name_hash}.png",
                    }
                )

            except Exception as e:
                Logger.instance().error(f"Error creating thumbnail: {str(e)}")
                return jsonify({"message": str(e)}), 500

        @self.app.route("/api/worksheet", methods=["GET"])
        def list_worksheets():
            """List all worksheets."""
            try:
                worksheets = self.worksheet_repo.list_worksheets()
                return jsonify(worksheets)
            except Exception as e:
                Logger.instance().error(f"Error listing worksheets: {str(e)}")
                return jsonify({"message": str(e)}), 500

        @self.app.route("/api/worksheet/<worksheet_id>", methods=["GET"])
        def get_worksheet(worksheet_id):
            """Get a specific worksheet."""
            try:
                worksheet = self.worksheet_repo.get_worksheet(worksheet_id)
                if worksheet is None:
                    return jsonify({"message": "Worksheet not found"}), 404
                return jsonify(worksheet)
            except Exception as e:
                Logger.instance().error(f"Error getting worksheet: {str(e)}")
                return jsonify({"message": str(e)}), 500

        @self.app.route("/api/worksheet", methods=["POST"])
        def create_worksheet():
            """Create a new worksheet."""
            try:
                data = request.get_json()
                if not data or "name" not in data:
                    return jsonify({"message": "Name is required"}), 400

                result = self.worksheet_repo.create_worksheet(
                    name=data["name"],
                    query=data.get("query", ""),
                    selected_source=data.get("selected_source"),
                )
                return jsonify(result), 201
            except Exception as e:
                Logger.instance().error(f"Error creating worksheet: {str(e)}")
                return jsonify({"message": str(e)}), 500

        @self.app.route("/api/worksheet/<worksheet_id>", methods=["PUT"])
        def update_worksheet(worksheet_id):
            """Update a worksheet."""
            try:
                data = request.get_json()
                if not data:
                    return jsonify({"message": "No update data provided"}), 400

                success = self.worksheet_repo.update_worksheet(worksheet_id, data)
                if not success:
                    return jsonify({"message": "Worksheet not found"}), 404

                return jsonify({"message": "Worksheet updated successfully"})
            except Exception as e:
                Logger.instance().error(f"Error updating worksheet: {str(e)}")
                return jsonify({"message": str(e)}), 500

        @self.app.route("/api/worksheet/<worksheet_id>", methods=["DELETE"])
        def delete_worksheet(worksheet_id):
            """Delete a worksheet."""
            try:
                success = self.worksheet_repo.delete_worksheet(worksheet_id)
                if not success:
                    return jsonify({"message": "Worksheet not found"}), 404
                return jsonify({"message": "Worksheet deleted successfully"})
            except Exception as e:
                Logger.instance().error(f"Error deleting worksheet: {str(e)}")
                return jsonify({"message": str(e)}), 500

        @self.app.route("/api/worksheet/session", methods=["GET"])
        def get_session_state():
            """Get the current session state."""
            try:
                worksheets = self.worksheet_repo.list_worksheets()
                return jsonify([w["session_state"] for w in worksheets])
            except Exception as e:
                Logger.instance().error(f"Error getting session state: {str(e)}")
                return jsonify({"message": str(e)}), 500

        @self.app.route("/api/worksheet/session", methods=["PUT"])
        def update_session_state():
            """Update the session state."""
            try:
                data = request.get_json()
                if not isinstance(data, list):
                    return jsonify({"message": "Expected array of session states"}), 400

                success = self.worksheet_repo.update_session_states(data)
                if not success:
                    return jsonify({"message": "Failed to update session state"}), 500

                return jsonify({"message": "Session state updated successfully"})
            except Exception as e:
                Logger.instance().error(f"Error updating session state: {str(e)}")
                return jsonify({"message": str(e)}), 500

        @self.app.route("/api/editors/installed", methods=["GET"])
        def get_installed_editors():
            editors, platform = get_editor_configs()

            # Check which editors are installed
            installed_editors = []
            for editor in editors:
                # Skip editors that don't support this platform
                if not editor["paths"][platform]:
                    continue

                for path in editor["paths"][platform]:
                    if os.path.exists(path):
                        # Only send back safe information to the client
                        installed_editors.append({"name": editor["name"], "id": editor["id"]})
                        break

            return jsonify(installed_editors)

        @self.app.route("/api/editors/open", methods=["POST"])
        def open_in_editor():
            data = request.get_json()
            if not data or "editorId" not in data or "filePath" not in data:
                return jsonify({"error": "Missing required parameters"}), 400

            editor_id = data["editorId"]
            file_path = data["filePath"]

            # Validate file path exists
            if not os.path.exists(file_path):
                return jsonify({"error": "File not found"}), 404

            # Get editor configurations
            editors, platform = get_editor_configs()

            # Find the selected editor
            editor_config = next((e for e in editors if e["id"] == editor_id), None)
            if not editor_config or not editor_config["commands"][platform]:
                return jsonify({"error": "Invalid editor for this platform"}), 400

            try:
                # Get the command for the selected editor
                command = editor_config["commands"][platform]

                # Special handling for VS Code on macOS
                if platform == "mac" and editor_id == "vscode":
                    if command[0] == "open":
                        # When using 'open' command, file path goes at the end
                        subprocess.Popen(command + ["--", file_path])
                    else:
                        # When using 'code' command directly
                        subprocess.Popen(command + [file_path])
                else:
                    # Normal handling for other editors
                    subprocess.Popen(command + [file_path])

                return jsonify({"message": "File opened successfully"}), 200
            except Exception as e:
                return jsonify({"error": str(e)}), 500

    @property
    def project(self):
        return self._project

    @project.setter
    def project(self, value):
        self._project_json = (
            Serializer(project=value).dereference().model_dump_json(exclude_none=True)
        )
        self._project = value



from typing import List
from pydantic import Field, PrivateAttr, ConfigDict, field_serializer, model_validator

from visivo.models.base.eval_string import EvalString
from visivo.models.destinations.fields import DestinationField
from visivo.models.test_run import TestRun
from visivo.models.base.named_model import NamedModel


"""
Tests allow you to assert on the computed values that are the output of a trace.  The tests are run with the `visivo test` command.

### Example
``` yaml
alerts:
    - name: Example Alert
      if: >{ anyTestFailed() && env.ENVIRONMENT == "PRODUCTION" }
      destinations: 
        - ${ ref(Production Slack) }
        - ${ ref(Production Email) }
```
"""


class Alert(NamedModel):
    model_config = ConfigDict(populate_by_name=True)
    if_: EvalString = Field(
        None,
        alias="if",
        description="A EvalString that must evaluate to true for the alert to fire",
    )
    destinations: List[DestinationField] = []

    _parent_test: str = PrivateAttr(default=None)

    @model_validator(mode="before")
    def rename_if(cls, values):
        if "if_" in values:
            values["if"] = values["if_"]
            del values["if_"]
        return values

    def set_parent_test(self, value: str):
        self._parent_test = value

    def alert(self, test_run: TestRun):
        for destination in self.destinations:
            destination.alert(test_run=test_run)



from typing import Literal

import click
from visivo.models.destinations.destination import Destination
from visivo.models.test_run import TestRun


class ConsoleDestination(Destination):
    called: bool = False
    message: str = "Console Destination Run"
    type: Literal["console"]

    def alert(self, test_run: TestRun):
        click.echo(self.message)
        self.called = True



import json
import re
from networkx import DiGraph, simple_cycles, is_directed_acyclic_graph
from visivo.models.dag import all_descendants_with_name, parse_filter_str
from typing import List, Optional


class ProjectDag(DiGraph):
    """
    Custom implementation of a DiGraph that adds additional methods for validation & data extraction.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._named_nodes_subgraph = None

    def get_named_nodes_subgraph(self):
        """Creates the named nodes subgraph if it doesn't exist"""
        if self._named_nodes_subgraph is None:
            self._named_nodes_subgraph = self.__compute_named_nodes_subgraph()
        return self._named_nodes_subgraph

    def validate_dag(self):
        if is_directed_acyclic_graph(self):
            return True

        circular_references = list(simple_cycles(self))
        if len(circular_references) > 0:
            circle = " -> ".join(list(map(lambda cr: cr.id(), circular_references[0])))
            circle += f" -> {circular_references[0][0].id()}."
            raise ValueError(f"Project contains a circular reference: {circle}")
        raise ValueError("Project is not a valid DAG.")

    def get_root_nodes(self):
        roots = [node for node, deg in self.in_degree() if deg == 0]
        return roots

    def get_node_by_path(self, path):
        for node in self.nodes():
            if hasattr(node, "path"):
                if node.path == path:
                    return node
        raise ValueError(f"Item with path {path} not found.")

    def get_node_by_name(self, name):
        for node in self.nodes():
            if hasattr(node, "name"):
                if node.name == name:
                    return node
        raise ValueError(f"Item with name {name} not found.")

    def get_nodes_by_types(self, types: List, is_named=Optional[bool]):
        nodes = []

        def node_match(node, type) -> bool:
            if not isinstance(node, type):
                return False
            if is_named is None:
                return True
            elif is_named == True:
                if hasattr(node, "name"):
                    return True
                else:
                    return False
            elif is_named == False:
                if not hasattr(node, "name"):
                    return True
                else:
                    return False
            return False

        for node in self.nodes():
            for type in types:
                if node_match(node, type):
                    nodes.append(node)
        return nodes

    def __compute_named_nodes_subgraph(self) -> "ProjectDag":
        """
        Creates a new DAG containing only named nodes, preserving direct relationships
        between named nodes even when connected through unnamed nodes.
        A direct relationship exists when there is a path between two named nodes
        that doesn't pass through any other named nodes.
        """
        named_nodes = []
        # First collect all named nodes
        for node in self.nodes():
            if hasattr(node, "name") and node.name is not None:
                named_nodes.append(node)

        # Create new DAG
        named_dag = ProjectDag()
        named_dag.add_nodes_from(named_nodes)

        # For each named node, find its direct named descendants
        for source in named_nodes:
            stack = [(source, child) for child in self.successors(source)]
            while stack:
                current_parent, current = stack.pop()
                if current in named_nodes and current != source:
                    # Found a named node, add edge from source to this node
                    named_dag.add_edge(source, current)
                elif current not in named_nodes:
                    # If unnamed node, continue searching its children
                    for child in self.successors(current):
                        stack.append((current_parent, child))
        project = named_dag.get_root_nodes()[0]
        named_dag.remove_node(project)
        return named_dag

    def get_named_children(self, node_name: str) -> List[str]:
        """
        Returns a list of names of all named parent nodes for the given node name.
        Uses the named nodes subgraph to determine relationships.
        """
        named_dag = self.get_named_nodes_subgraph()

        try:
            node = named_dag.get_node_by_name(node_name)
            nodes = []
            for predecessor in named_dag.predecessors(node):
                if hasattr(predecessor, "name") and predecessor.name is not None:
                    nodes.append(predecessor.name)
            return nodes
        except ValueError:
            return []

    def get_named_parents(self, node_name: str) -> List[str]:
        """
        Returns a list of names of all named child nodes for the given node name.
        Uses the named nodes subgraph to determine relationships.
        """
        named_dag = self.get_named_nodes_subgraph()

        try:
            node = named_dag.get_node_by_name(node_name)
            nodes = []
            for successor in named_dag.successors(node):
                if hasattr(successor, "name") and successor.name is not None:
                    nodes.append(successor.name)
            return nodes
        except ValueError:
            return []

    def filter_dag(self, filter_str) -> List["ProjectDag"]:
        from networkx import (
            subgraph,
            shortest_path_length,
            descendants,
            ancestors,
            compose,
        )

        if not filter_str:
            return [self]

        filtered_dags = []
        filters = parse_filter_str(filter_str)
        for filter in filters:
            post, name, pre = filter
            item = all_descendants_with_name(name=name, dag=self)
            if len(item) == 1:
                item = item[0]
            else:
                continue
            pre_length = 0
            post_length = 0
            a = ancestors(self, item)
            d = descendants(self, item)
            if pre == "+":
                pre_length = len(a)
            elif pre:
                pre_length = int(pre.replace("+", ""))
            if post == "+":
                post_length = len(d)
            elif post:
                post_length = int(post.replace("+", ""))

            def matches_length_and_side(node):
                return (
                    (node in a and shortest_path_length(self, node, item) <= pre_length)
                    or (node in d and shortest_path_length(self, item, node) <= post_length)
                    or node == item
                )

            filtered_nodes = [node for node in self.nodes if matches_length_and_side(node)]

            filtered_dags.append(subgraph(self, filtered_nodes))

        def combine_dags(dags):
            combined_dags = []
            while dags:
                dag = dags.pop()
                combined = False
                for i, combined_dag in enumerate(combined_dags):
                    if combined_dag.nodes & dag.nodes:
                        combined_dags[i] = compose(dag, combined_dag)
                        combined = True
                        break
                if not combined:
                    combined_dags.append(dag)
            return combined_dags

        combined_dags = filtered_dags.copy()
        while True:
            len_before = len(combined_dags)
            combined_dags = combine_dags(combined_dags)
            if len_before == len(combined_dags):
                break
        return combined_dags

    def get_diff_dag_filter(self, existing_project, existing_dag_filter):
        """
        Compares this project DAG with the existing project's DAG filtered by the existing filter.
        It identifies nodes that have changed or are new and returns a filter string that includes these nodes.

        Parameters:
        - existing_project (Project): The existing project to compare with.
        - existing_dag_filter (str): The filter string used to filter the existing project's DAG.

        Returns:
        - str: A comma-separated filter string that selects all nodes that are dependent on the changed nodes.
        """
        existing_dags = (
            existing_project.dag().get_named_nodes_subgraph().filter_dag(existing_dag_filter)
        )
        existing_nodes = [node for dag in existing_dags for node in dag.nodes()]
        new_dags = self.get_named_nodes_subgraph().filter_dag(existing_dag_filter)
        new_nodes = [node for dag in new_dags for node in dag.nodes()]
        changed_dag_filter = []
        for new_node in new_nodes:
            existing_node = next((n for n in existing_nodes if n.name == new_node.name), None)
            if existing_node:
                if json.dumps(existing_node.model_dump_json(), sort_keys=True) != json.dumps(
                    new_node.model_dump_json(), sort_keys=True
                ):
                    changed_dag_filter.append(f"{new_node.name}+")
            else:
                changed_dag_filter.append(f"{new_node.name}+")

        return ",".join(changed_dag_filter)



from time import time

start_time = time()
from visivo.logger.logger import Logger, TypeEnum

Logger.instance().info("Starting Visivo...")
import click
import os
from dotenv import load_dotenv
from pydantic import ValidationError
import sys

from visivo.parsers.line_validation_error import LineValidationError

from visivo.commands.dbt import dbt
from visivo.commands.deploy import deploy
from visivo.commands.serve import serve
from visivo.commands.run import run
from visivo.commands.dist import dist
from visivo.commands.compile import compile
from visivo.commands.init import init
from visivo.commands.create import create
from visivo.commands.test import test
from visivo.commands.aggregate import aggregate
from visivo.commands.archive import archive
from visivo.commands.authorize import authorize
from visivo.commands.list import list
from visivo.version import VISIVO_VERSION


@click.group()
@click.option("-p", "--profile", is_flag=True)
@click.option("-e", "--env-file", default=".env")
@click.version_option(version=VISIVO_VERSION)
def visivo(env_file, profile):
    Logger.instance().set_type(TypeEnum.spinner)
    load_env(env_file)

    # Profiling can be done with https://github.com/nschloe/tuna
    #  `tuna visivo-profile.dmp`
    # If you need to profile the import time, you can use the following command:
    #  `python -X importtime -m visivo.command_line compile 2> import.log`
    #  `tuna import.log`
    if profile:
        import cProfile
        import atexit

        Logger.instance().info("Profiling...")
        pr = cProfile.Profile()
        pr.enable()

        def exit():
            pr.disable()
            Logger.instance().info("Profiling completed")
            pr.dump_stats("visivo-profile.dmp")

        atexit.register(exit)


visivo.add_command(init)
visivo.add_command(dbt)
visivo.add_command(compile)
visivo.add_command(run)
visivo.add_command(serve)
visivo.add_command(deploy)
visivo.add_command(dist)
visivo.add_command(test)
visivo.add_command(aggregate)
visivo.add_command(archive)
visivo.add_command(authorize)
visivo.add_command(create)
visivo.add_command(list)


def load_env(env_file):
    if os.path.isfile(env_file):
        Logger.instance().debug(f"Loading env file: {env_file}")
        load_dotenv(env_file)


def print_issue_url():
    import traceback
    import urllib.parse

    stack_trace = "".join(traceback.format_exc())
    command = " ".join(sys.argv)
    issue_body = f"Command: {command}\n\nStack Trace:\n```\n{stack_trace}\n```"
    encoded_body = urllib.parse.quote(issue_body)
    issue_url = f"https://github.com/visivo-io/visivo/issues/new?body={encoded_body}"

    Logger.instance().error(
        f"\x1b]8;;{issue_url}\x1b\\Click here to report this issue\x1b]8;;\x1b\\"
    )


def safe_visivo():
    try:
        visivo(standalone_mode=False)
        Logger.instance().info(f"Visivo execution time: {round(time() - start_time, 2)}s")
    except (ValidationError, LineValidationError) as e:
        Logger.instance().error(str(e))
        sys.exit(1)
    except Exception as e:
        if "STACKTRACE" in os.environ and os.environ["STACKTRACE"] == "true":
            raise e
        Logger.instance().error("An unexpected error has occurred")
        Logger.instance().error(str(e))
        Logger.instance().error(
            "To print more error information add the 'STACKTRACE=true' environment variable."
        )
        print_issue_url()
        sys.exit(1)


if __name__ == "__main__":
    safe_visivo()



from visivo.models.base.query_string import QUERY_STRING_VALUE_REGEX, QueryString
from visivo.models.base.base_model import STATEMENT_REGEX
from typing import Annotated, Any, Union
import re
from pydantic import Discriminator, Tag


def get_discriminator_value(value: Any) -> str:
    if isinstance(value, str) and re.match(QUERY_STRING_VALUE_REGEX, value):
        return "Query"
    elif isinstance(value, QueryString):
        return "Query"

    return "String"


QueryOrStringField = Annotated[
    Union[
        Annotated[QueryString, Tag("Query")],
        Annotated[str, Tag("String")],
    ],
    Discriminator(get_discriminator_value),
]



from visivo.jobs.dag_runner import DagRunner
from visivo.models.project import Project


class FilteredRunner:
    def __init__(
        self,
        project: Project,
        output_dir: str,
        threads: int = 8,
        soft_failure: bool = False,
        dag_filter: str = None,
        server_url: str = None,
    ):
        self.project = project
        self.output_dir = output_dir
        self.threads = threads
        self.soft_failure = soft_failure
        self.dag_filter = dag_filter
        self.server_url = server_url
        self.project_dag = project.dag()

    def run(self):
        for job_dag in self.project_dag.filter_dag(self.dag_filter):
            dag_runner = DagRunner(
                project=self.project,
                output_dir=self.output_dir,
                threads=self.threads,
                soft_failure=self.soft_failure,
                server_url=self.server_url,
                job_dag=job_dag,
            )
            dag_runner.run()



import click
from visivo.commands.options import (
    dist_dir,
    output_dir,
)


@click.command()
@output_dir
@dist_dir
def dist(output_dir, dist_dir):
    """
    Creates a distributable version of this dashboard and stores it in a 'dist' folder.
    This folder can be statically deployed to any web server, hosting service, or bucket storage.

    Note: Make sure to run `visivo run` before running `visivo dist`.
    """
    from visivo.logger.logger import Logger

    Logger.instance().debug("Creating Dist")

    from visivo.commands.dist_phase import dist_phase

    dist_phase(
        output_dir=output_dir,
        dist_dir=dist_dir,
    )
    Logger.instance().success(f"Created dist folder: {dist_dir}")



import os
import polars as pl
import json
import io


class Aggregator:
    @classmethod
    def aggregate(cls, json_file: str, trace_dir: str):
        data_frame = pl.read_json(json_file)
        data_frame = data_frame.rename({col: col.replace("|", ".") for col in data_frame.columns})
        cls.aggregate_data_frame(data_frame=data_frame, trace_dir=trace_dir)

    @classmethod
    def aggregate_data_frame(cls, data_frame: pl.DataFrame, trace_dir: str):
        data_frame = data_frame.rename({col: col.replace("|", ".") for col in data_frame.columns})
        agg_exprs = [
            pl.col(col).implode().alias(col) for col in data_frame.columns if col != "cohort_on"
        ]
        aggregated = data_frame.group_by("cohort_on").agg(agg_exprs)
        # Write JSON to a string buffer
        buf = io.StringIO()
        aggregated.write_json(buf)
        buf.seek(0)
        rows = json.load(buf)
        # Transform to pandas-style dict
        result = {}
        for row in rows:
            cohort = row["cohort_on"]
            value = {}
            for k, v in row.items():
                if k == "cohort_on":
                    continue
                # Unwrap single-item lists if the item is itself a list
                if isinstance(v, list) and len(v) == 1 and isinstance(v[0], list):
                    value[k] = v[0]
                else:
                    value[k] = v
            result[cohort] = value
        os.makedirs(trace_dir, exist_ok=True)
        with open(f"{trace_dir}/data.json", "w") as fp:
            json.dump(result, fp, indent=4)



from visivo.models.base.named_model import NamedModel
from visivo.models.test_run import TestRun


class Destination(NamedModel):
    def alert(self, test_run: TestRun):
        raise NotImplementedError("Please Implement this method")



from jsonschema_rs import ValidationError
from pydantic import ConfigDict, BaseModel
from typing import Any, Dict
import json


def get_message_from_error(error: ValidationError, schema: Dict[str, Any]) -> str:
    message = str(error.message)
    if "is not valid under any of the schemas listed in the" in message:
        current = schema["properties"]
        for part in error.instance_path:
            current = current.get(part, {})
        message_parts = []
        for oneOf in current.get("oneOf", []):
            message_parts.append(json.dumps(oneOf))
        message = "Value does not match any of the following schemas: \n" + "\n or \n".join(
            message_parts
        )
    return message


class JsonSchemaBase(BaseModel):

    def dict(self, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        kwargs.setdefault("exclude_none", True)
        return super().model_dump(*args, **kwargs)

    model_config = ConfigDict(extra="allow")



import click


@click.command()
@click.option("--project-name", type=str, help="The name of the project to initialize")
def create(project_name):
    """
    Enables a quick set up by writing your source & api credentials to an env file.
    """
    from visivo.logger.logger import Logger

    Logger.instance().debug("Create")

    from visivo.commands.create_phase import create_phase

    create_phase(project_name)
    Logger.instance().success("Done")



import os
import platform
import click
from threading import Thread
from visivo.logger.logger import Logger
import urllib.parse
from visivo.commands.options import host
from visivo.tokens.token_functions import (
    get_existing_token,
)
from visivo.tokens.web_utils import open_url
from visivo.tokens.server import (
    run_flask_server,
    token_received_event,
    FLASK_PORT,
)

PROFILE_PATH = os.path.expanduser("~/.visivo/profile.yml")
CALLBACK_RESPONSE_WAIT_TIME = 120


@click.command()
@host
def authorize(host):
    """
    Handles the authorization process.

    Checks for an existing token, and if the user chooses to add a
    new token, starts a Flask server (listening for a callback on
    http://localhost:5001/authorize-device-token). It then opens an external webapp
    (at <host>/authorize-device) for testing. The external webapp should eventually
    send a callback to our Flask server.

    The process will wait for up to CALLBACK_RESPONSE_WAIT_TIME seconds for the callback.
    If the callback is not received, you will be prompted to cancel or continue waiting.
    """
    Logger.instance().spinner.stop()

    existing_token = get_existing_token()
    if existing_token:
        Logger.instance().info(f"A token already exists in your profile: {existing_token}")
        if not click.confirm("Do you want to add a new token?"):
            Logger.instance().info("Authorization cancelled. Using the existing token.")
            return

    server_thread = Thread(
        target=run_flask_server,
        kwargs={"port": FLASK_PORT, "base_url": host},
        daemon=True,
    )
    server_thread.start()

    device_name = platform.node()
    redirect_url = f"http://localhost:{FLASK_PORT}/authorize-device-token"

    params = {"redirect_url": redirect_url, "name": device_name}

    query_string = urllib.parse.urlencode(params)

    full_url = f"{host}/authorize-device?{query_string}"

    Logger.instance().info(f"Attempting to open URL: {full_url}")
    webbrowser_opened = open_url(full_url)

    if not webbrowser_opened:
        Logger.instance().error(
            "We couldn't automatically open the external web app. "
            "If you have multiple browser windows open, please copy the URL below and paste it "
            "into the address bar of the browser window where you're logged into Visivo."
        )
        Logger.instance().info(full_url)

    Logger.instance().info("Waiting for visivo token response (timeout: 120s)...")
    Logger.instance().info(
        "Please copy and paste the above URL into your logged in visivo browser window if this did not open automatically or if it opened a not authenticated browser."
    )

    try:
        if not token_received_event.wait(CALLBACK_RESPONSE_WAIT_TIME):
            Logger.instance().error("Timeout reached while waiting for callback.")
            if click.confirm("Do you want to cancel the authorization process?"):
                Logger.instance().info(
                    "Authorization cancelled. No token received. To add a token, try running the 'visivo authorize' command again"
                )
                return
            else:
                Logger.instance().info("Continuing to wait for token. Press Ctrl+C to cancel.")
                token_received_event.wait()  # Wait indefinitely.
    except KeyboardInterrupt:
        Logger.instance().info("Authorization process cancelled by user.")
        return
